import pytest
from influxable import InfluxDBApi, exceptions
from influxable.connection import Connection


class TestConnection:
    def check_if_connection_reached(self, connection):
        query = 'SHOW DATABASES'
        InfluxDBApi.execute_query(connection.request, query)

    def test_create_instance_with_no_args_success(self):
        connection = Connection()
        self.check_if_connection_reached(connection)
        assert connection is not None

    def test_create_instance_with_args_success(self):
        connection = Connection(
            user='admin',
            password='changeme'
        )
        self.check_if_connection_reached(connection)
        assert connection is not None

    def test_create_instance_with_bad_url_fail(self):
        with pytest.raises(exceptions.InfluxDBInvalidURLError):
            Connection(base_url="incorrect_url")

    def test_create_instance_with_bad_database_name_fail(self):
        pytest.skip()
        with pytest.raises(exceptions.InfluxDBInvalidURLError):
            Connection(database_name="invalid_database_name")



import arrow
import random
from influxable import Influxable, attributes
from influxable.measurement import Measurement

client = Influxable.get_instance()


def get_populate_measurement(name):
    class PopulateMeasurement(Measurement):
        measurement_name = name

        time = attributes.TimestampFieldAttribute(precision='s')
        phase = attributes.TagFieldAttribute()
        value = attributes.FloatFieldAttribute()
    return PopulateMeasurement


class Populate():
    @staticmethod
    def run(*args, **kwargs):
        measurement_name = kwargs.get('measurement_name')
        PopulateMeasurement = get_populate_measurement(measurement_name)

        start_date = kwargs.get('start_date')
        end_date = kwargs.get('end_date')
        interval_delay = kwargs.get('interval_delay')

        tags = kwargs.get('tags')
        max_count = kwargs.get('max_count_of_values', None)

        min_value = kwargs.get('min_value')
        max_value = kwargs.get('max_value')
        value_range = (min_value, max_value)
        date_range = list(arrow.Arrow.span_range('month', start_date, end_date))
        date_range.reverse()

        count = 0
        for current_start_month, current_end_month in date_range:
            if max_count is not None and count > max_count:
                break

            points = []
            current_date = current_end_month
            while current_date >= current_start_month:

                if max_count is not None and count > max_count:
                    break
                points.append(PopulateMeasurement(
                    phase=random.choice(tags),
                    value=random.uniform(*value_range),
                    time=current_date.timestamp,
                ))
                current_date = current_date.shift(minutes=-interval_delay)
                count += 1
            PopulateMeasurement.bulk_save(points)



import os
from pathlib import Path
from jinja2 import Environment, FileSystemLoader
from .attributes import BaseAttribute, GenericFieldAttribute, \
    TagFieldAttribute, TimestampFieldAttribute
from .db.query import Query, BulkInsertQuery
from .response import InfluxDBResponse
from .serializers import MeasurementPointSerializer
from .exceptions import InfluxDBAttributeValueError

EXTENDED_ATTRIBUTE_PREFIX_NAME = '__attribute__'

TEMPLATE_FILE_NAME = 'simple_measurement.py.jinja'


class MeasurementMeta(type):
    def __init__(cls, name, *args, **kwargs):
        super(MeasurementMeta, cls).__init__(name, *args, **kwargs)
        attribute_names = cls._get_attribute_names()
        cls._extend_attributes(attribute_names)

        get_query = cls._factory_get_query()
        setattr(cls, 'get_query', get_query)

    def __call__(cls, *args, **kwargs):
        setattr(cls, '_get_attributes', cls._get_attributes)
        instance = type.__call__(cls, *args, **kwargs)
        return instance

    def _factory_get_query(cls):
        def get_query():
            class MeasurementQuery(Query):
                def __init__(self):
                    super(MeasurementQuery, self).__init__()

                def format(self, result, parser_class=cls.parser_class):
                    return parser_class(result, cls).convert()

                def evaluate(self, parser_class=cls.parser_class):
                    result = InfluxDBResponse(self.execute())
                    result.raise_if_error()
                    formatted_result = self.format(result, parser_class)
                    return formatted_result
            return MeasurementQuery().from_measurements(cls.measurement_name)
        return get_query

    def _get_attribute_names(cls):
        def filter_func(x):
            return isinstance(variables[x], BaseAttribute)
        variables = cls.__dict__
        attribute_names = list(filter(filter_func, variables))
        return attribute_names

    def _get_attributes(cls):
        def filter_func(x):
            return isinstance(x, BaseAttribute)
        variables = cls.__dict__.values()
        attributes = list(filter(filter_func, variables))
        return attributes

    def _get_timestamp_attributes(cls):
        def filter_func(x):
            return isinstance(x, TimestampFieldAttribute)
        attributes = cls._get_attributes()
        timestamp_attributes = list(filter(filter_func, attributes))
        return timestamp_attributes

    def _extend_attributes(cls, attribute_names):
        def generate_getter_and_setter(attr_name):
            def getx(self):
                attribute_field = getattr(self, attr_name)
                return attribute_field.get_internal_value()

            def setx(self, value):
                attribute_field = getattr(self, attr_name)
                attribute_field.set_internal_value(value)
            return getx, setx

        for attribute_name in attribute_names:
            ext_attribute_name = EXTENDED_ATTRIBUTE_PREFIX_NAME + attribute_name
            attribute_field = getattr(cls, attribute_name)
            attribute_field.attribute_name = attribute_name
            attribute_field.ext_attribute_name = ext_attribute_name

            getx, setx = generate_getter_and_setter(ext_attribute_name)
            prop = property(getx, setx)
            setattr(cls, ext_attribute_name, attribute_field)
            setattr(cls, attribute_name, prop)


class Measurement(object, metaclass=MeasurementMeta):
    parser_class = MeasurementPointSerializer
    measurement_name = 'default'

    def __init__(self, **kwargs):
        self.check_attribute_values(**kwargs)
        self.clone_attributes()
        self.fill_values(**kwargs)

    def check_attribute_values(self, **kwargs):
        def filter_required_attributes(x):
            return not x.default and not x.is_nullable

        attributes = self._get_attributes()
        required_attributes = list(filter(
            filter_required_attributes,
            attributes,
        ))
        required_attributes_names = [
            r.attribute_name
            for r in required_attributes
        ]
        for key in required_attributes_names:
            if key not in kwargs:
                raise InfluxDBAttributeValueError(
                    'The attribute \'{}\' cannot be nullable'.format(key)
                )

    def clone_attributes(self):
        attributes = self._get_attributes()
        for attr in attributes:
            cloned_attribute = attr.clone()
            cloned_attribute.attribute_name = attr.attribute_name
            cloned_attribute.ext_attribute_name = attr.ext_attribute_name
            setattr(self, attr.ext_attribute_name, cloned_attribute)

    def dict(self):
        dict_values = {}
        attributes = self.get_attributes()
        for attr in attributes:
            dict_values[attr.attribute_name] = getattr(self, attr.attribute_name)
        return dict_values

    def get_attributes(self):
        def filter_func(x):
            return isinstance(x, BaseAttribute)
        variables = self.__dict__.values()
        attributes = list(filter(filter_func, variables))
        return attributes

    def get_attribute_names(self):
        attributes = self.get_attributes()
        attribute_names = [attr.attribute_name for attr in attributes]
        return attribute_names

    def get_ext_attribute_names(self):
        attributes = self.get_attributes()
        attribute_names = [attr.ext_attribute_name for attr in attributes]
        return attribute_names

    def get_timestamp_attributes(self):
        def filter_func(x):
            return isinstance(x, TimestampFieldAttribute)
        attributes = self.get_attributes()
        timestamp_attributes = list(filter(filter_func, attributes))
        return timestamp_attributes

    def get_prep_value(self):
        def factory_filter_func(cls):
            def filter_func(attr):
                return isinstance(attr, cls)
            return filter_func

        attributes = self.get_attributes()
        field_attributes = list(filter(
            factory_filter_func(GenericFieldAttribute),
            attributes,
        ))
        tag_attributes = list(filter(
            factory_filter_func(TagFieldAttribute),
            attributes,
        ))
        timestamp_attributes = list(filter(
            factory_filter_func(TimestampFieldAttribute),
            attributes,
        ))

        prep_value_groups = []
        attributes_groups = [tag_attributes, field_attributes, timestamp_attributes]
        for attr_group in attributes_groups:
            prep_value_group = []
            for attr in attr_group:
                if attr.raw_value is not None:
                    attr_prep_value = attr.get_prep_value()
                    attr_name = attr.name or attr.attribute_name
                    if not isinstance(attr, TimestampFieldAttribute):
                        prep_value = '{}={}'.format(attr_name, attr_prep_value)
                    else:
                        prep_value = '{}'.format(attr_prep_value)
                    prep_value_group.append(prep_value)
            str_prep_value_group = ','.join(prep_value_group)
            prep_value_groups.append(str_prep_value_group)

        if prep_value_groups[0]:
            prep_value_groups[0] = ','.join(
                [self.measurement_name] + [prep_value_groups[0]]
            )
        else:
            prep_value_groups[0] = self.measurement_name
        final_prep_value = ' '.join(prep_value_groups)
        return final_prep_value

    def fill_values(self, **kwargs):
        try:
            for key, value in kwargs.items():
                setattr(self, key, value)
        except Exception as err:
            print('key', key)
            msg = '<\'{key}\'> : {msg}'.format(key=key, msg=err)
            raise InfluxDBAttributeValueError(msg)

    def items(self):
        return self.dict().items()

    @staticmethod
    def bulk_save(points):
        if not isinstance(points, list):
            raise InfluxDBAttributeValueError('points must be a list')
        str_points = ''
        for point in points:
            if not isinstance(point, Measurement):
                raise InfluxDBAttributeValueError(
                    'type of point must be Measurement'
                )
            prep_value = point.get_prep_value()
            str_points += prep_value
            str_points += '\n'
        return BulkInsertQuery(str_points).execute()


def SimpleMeasurement(measurement_name, field_names, tag_names=[]):
    current_dir_path = os.path.dirname(os.path.realpath(__file__))
    template_folder_path = Path(current_dir_path) / './templates/'
    template_folder_path = template_folder_path.resolve().as_posix()

    j2_env = Environment(
        loader=FileSystemLoader(template_folder_path),
        trim_blocks=True,
    )
    rendered_template = j2_env\
        .get_template(TEMPLATE_FILE_NAME)\
        .render(
            measurement_name=measurement_name,
            field_names=field_names,
            tag_names=tag_names,
        )

    exec(rendered_template)
    anonymous_measurement = locals().get('AnonymousMeasurement')
    return anonymous_measurement



from .admin import InfluxDBAdmin
from .criteria import Field
from .query import Query, RawQuery, BulkInsertQuery


__all__ = [
    'InfluxDBAdmin',
    'Field',
    'Query',
    'RawQuery',
    'BulkInsertQuery',
]



import os
import humps
from pathlib import Path
from influxable import Influxable, attributes
from influxable.db import InfluxDBAdmin
from jinja2 import Environment, FileSystemLoader

client = Influxable()

TEMPLATE_FILE_NAME = 'auto_generated_measurement.py.jinja'


def get_classname(cls):
    return cls.__name__


def pascalize(str):
    return humps.pascalize(str)


FIELD_TYPE_ATTRIBUTE_MAP = {
    'integer': get_classname(attributes.IntegerFieldAttribute),
    'float': get_classname(attributes.FloatFieldAttribute),
    'boolean': get_classname(attributes.BooleanFieldAttribute),
    'string': get_classname(attributes.StringFieldAttribute),
}

TAG_CLASSNAME = get_classname(attributes.TagFieldAttribute)


class AutoGenerateMeasurement():
    @staticmethod
    def get_template_path():
        current_dir_path = os.path.dirname(os.path.realpath(__file__))
        template_folder_path = Path(current_dir_path) / '../templates/'
        return template_folder_path.resolve().as_posix()

    @staticmethod
    def retrieve_measurement_fields():
        all_measurements_fields = {}
        measurements_field_keys = InfluxDBAdmin.show_field_keys()
        measurements_tag_keys = InfluxDBAdmin.show_tag_keys()

        for measurement_field_keys in measurements_field_keys:
            measurement_name, fields = list(measurement_field_keys.items())[0]
            all_measurements_fields.setdefault(measurement_name, [])
            for field in fields:
                field_key = field['fieldKey']
                field_type = field['fieldType']
                field_tuple = (field_key, FIELD_TYPE_ATTRIBUTE_MAP[field_type])
                all_measurements_fields[measurement_name].append(field_tuple)

        for measurement_tag_keys in measurements_tag_keys:
            measurement_name, tags = list(measurement_tag_keys.items())[0]
            all_measurements_fields.setdefault(measurement_name, [])
            for tag in tags:
                tag_key = tag['tagKey']
                tag_tuple = (tag_key, TAG_CLASSNAME)
                all_measurements_fields[measurement_name].append(tag_tuple)

        return all_measurements_fields

    @staticmethod
    def run(*args, **kwargs):
        output_file_name = kwargs.get('output_file_name')
        measurements_fields = AutoGenerateMeasurement.retrieve_measurement_fields()
        template_path = AutoGenerateMeasurement.get_template_path()

        j2_env = Environment(
            loader=FileSystemLoader(template_path),
            trim_blocks=True,
        )

        j2_env.filters['pascalize'] = pascalize

        rendered_template = j2_env\
            .get_template(TEMPLATE_FILE_NAME)\
            .render(measurements_fields=measurements_fields)

        with open(output_file_name, "wb") as f:
            f.write(str.encode(rendered_template))



import pytest
import requests
from influxable import Influxable, InfluxDBApi, exceptions


class TestInfluxApi:
    def get_instance(self):
        instance = Influxable.get_instance()
        return instance

    def test_get_debug_requests_success(self):
        instance = self.get_instance()
        res = InfluxDBApi.get_debug_requests(
            instance.connection.request,
            seconds=10,
        )
        assert res is not None

    def test_get_debug_requests_silent_fail(self):
        instance = self.get_instance()
        res = InfluxDBApi.get_debug_requests(
            instance.connection.request,
            seconds='k',
        )
        assert res is not None

    def test_get_debug_vars_success(self):
        instance = self.get_instance()
        res = InfluxDBApi.get_debug_vars(instance.connection.request)
        assert res is not None

    def test_ping_success(self):
        instance = self.get_instance()
        res = InfluxDBApi.ping(instance.connection.request)
        assert res is True

    def test_ping_verbose_success(self):
        instance = self.get_instance()
        res = InfluxDBApi.ping(instance.connection.request, verbose=True)
        assert res is not None
        assert 'version' in res

    def test_ping_verbose_silent_fail(self):
        instance = self.get_instance()
        res = InfluxDBApi.ping(instance.connection.request, verbose='k')
        assert res is True

    def test_execute_query_success(self):
        query = 'SHOW DATABASES'
        instance = self.get_instance()
        res = InfluxDBApi.execute_query(instance.connection.request, query)
        assert res is not None
        assert 'results' in res

    def test_execute_query_bad_query_fail(self):
        with pytest.raises(exceptions.InfluxDBBadQueryError):
            query = 'SELECT *'
            instance = self.get_instance()
            InfluxDBApi.execute_query(instance.connection.request, query)

    def test_execute_query_bad_method_fail(self):
        with pytest.raises(requests.exceptions.HTTPError):
            query = 'SHOW DATABASES'
            instance = self.get_instance()
            InfluxDBApi.execute_query(
                instance.connection.request,
                query,
                method='delete',
            )

    def test_write_points_success(self):
        precision = 's'
        points = 'mymeas,mytag=1 myfield=90 1463683075'
        instance = self.get_instance()
        res = InfluxDBApi.write_points(
            instance.connection.request,
            points,
            precision,
        )
        assert res is True

    def test_write_points_without_ts_success(self):
        precision = 's'
        points = 'mymeas,mytag=2 myfield=91'
        instance = self.get_instance()
        res = InfluxDBApi.write_points(
            instance.connection.request,
            points,
            precision,
        )
        assert res is True

    def test_write_points_multiple_success(self):
        precision = 's'
        points = '''
        mymeas,mytag=1 myfield=90 1463683075
        mymeas,mytag=3 myfield=34 1463683025
        '''
        instance = self.get_instance()
        res = InfluxDBApi.write_points(
            instance.connection.request,
            points,
            precision,
        )
        assert res is True

    def test_write_points_with_policy_success(self):
        rp = 'myrp'
        points = 'mymeas,mytag=1 myfield=90'
        instance = self.get_instance()
        res = InfluxDBApi.write_points(
            instance.connection.request,
            points,
            retention_policy_name=rp,
        )
        assert res is True

    def test_write_points_bad_timestamp_fail(self):
        with pytest.raises(exceptions.InfluxDBInvalidTimestampError):
            points = 'mymeas,mytag=1 myfield=90 invalid'
            instance = self.get_instance()
            InfluxDBApi.write_points(
                instance.connection.request,
                points,
            )



import pytest
from influxable.db.criteria import Field
from influxable.db.query import RawQuery, Query, BulkInsertQuery
from influxable.db.function.transformations import Abs
from influxable import exceptions


class TestDBRawQuery:
    def test_execute_query_success(self):
        str_query = 'SHOW DATABASES'
        raw_query = RawQuery(str_query)
        assert raw_query.query == str_query
        res = raw_query.execute()
        assert 'results' in res

    def test_execute_query_empty_query_fail(self):
        with pytest.raises(exceptions.InfluxDBEmptyRequestError):
            str_query = ''
            raw_query = RawQuery(str_query)
            raw_query.execute()

    def test_execute_query_bad_query_fail(self):
        with pytest.raises(exceptions.InfluxDBBadQueryError):
            str_query = 'SELECT *'
            raw_query = RawQuery(str_query)
            raw_query.execute()


class TestDBSelectQuery:
    def test_all_fields_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT * FROM "default"'
        res = query.execute()
        assert 'results' in res

    def test_one_field_success(self):
        query = Query()\
            .select('field1')\
            .from_measurements('default')
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT field1 FROM "default"'
        res = query.execute()
        assert 'results' in res

    def test_two_field_success(self):
        query = Query()\
            .select('field1', 'field2')\
            .from_measurements('default')
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT field1,field2 FROM "default"'
        res = query.execute()
        assert 'results' in res

    def test_with_function_success(self):
        query = Query()\
            .select(Abs('field1'))\
            .from_measurements('default')
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT ABS(field1) FROM "default"'
        res = query.execute()
        assert 'results' in res

    def test_chain_success(self):
        query = Query()\
            .select('field1')\
            .select('field2')\
            .from_measurements('default')
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT field2 FROM "default"'
        res = query.execute()
        assert 'results' in res

    def test_none_field_fail(self):
        with pytest.raises(exceptions.InfluxDBInvalidTypeError):
            Query()\
                .select(None)\
                .from_measurements('default')


class TestDBIntoQuery:
    def test_one_measurement_success(self):
        query = Query()\
            .select()\
            .from_measurements('cpu')\
            .into('cpu_copy')
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT * INTO cpu_copy FROM "cpu"'
        res = query.execute()
        assert 'results' in res

    def test_chain_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .into('cpu_copy')\
            .into('cpu_copy_2')
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT * INTO cpu_copy_2 FROM "default"'
        res = query.execute()
        assert 'results' in res

    def test_with_bad_measurement_fail(self):
        with pytest.raises(exceptions.InfluxDBInvalidTypeError):
            Query()\
                .select()\
                .from_measurements('cpu')\
                .into(True)


class TestDBFromQuery:
    def test_one_measurement_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT * FROM "default"'
        res = query.execute()
        assert 'results' in res

    def test_two_measurement_success(self):
        query = Query()\
            .select()\
            .from_measurements('default', 'default_2')
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT * FROM "default","default_2"'
        res = query.execute()
        assert 'results' in res

    def test_chain_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .from_measurements('default2')
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT * FROM "default2"'
        res = query.execute()
        assert 'results' in res

    def test_with_empty_measurement_fail(self):
        with pytest.raises(exceptions.InfluxDBInvalidTypeError):
            Query()\
                .select()\
                .from_measurements()

    def test_with_bad_measurement_fail(self):
        with pytest.raises(exceptions.InfluxDBInvalidTypeError):
            Query()\
                .select()\
                .from_measurements(True)


class TestDBWhereQuery:
    def test_one_criteria_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .where(
                Field('value') > 800,
            )
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT * FROM "default" WHERE "value" > 800'
        res = query.execute()
        assert 'results' in res

    def test_two_criteria_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .where(
                Field('value') < 400,
                Field('value') > 800,
            )
        prepared_query = query._get_prepared_query()
        test_query = 'SELECT * FROM "default" WHERE "value" < 400 AND "value" > 800'
        assert prepared_query == test_query
        res = query.execute()
        assert 'results' in res

    def test_with_function_success(self):
        pytest.skip()
        query = Query()\
            .select()\
            .from_measurements('default')\
            .where(
                Field(Abs('value')) < 400,
            )
        prepared_query = query._get_prepared_query()
        test_query = 'SELECT * FROM "default" WHERE "value" < 400 AND "value" > 800'
        assert prepared_query == test_query
        res = query.execute()
        assert 'results' in res

    def test_chain_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .where(
                Field('value') < 400,
            )\
            .where(
                Field('value') < 500,
            )
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT * FROM "default" WHERE "value" < 500'
        res = query.execute()
        assert 'results' in res

    def test_with_empty_criteria_fail(self):
        with pytest.raises(exceptions.InfluxDBInvalidTypeError):
            Query()\
                .select()\
                .from_measurements('default')\
                .where()

    def test_with_bad_criteria_fail(self):
        with pytest.raises(exceptions.InfluxDBInvalidTypeError):
            Query()\
                .select()\
                .from_measurements('default')\
                .where(True)


class TestDBLimitQuery:
    def test_with_good_value_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .limit(10)
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT * FROM "default" LIMIT 10'
        res = query.execute()
        assert 'results' in res

    def test_chain_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .limit(10)\
            .limit(100)
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT * FROM "default" LIMIT 100'
        res = query.execute()
        assert 'results' in res

    def test_with_empty_fail(self):
        with pytest.raises(TypeError):
            Query()\
                .select()\
                .from_measurements('default')\
                .limit()

    def test_with_negative_value_fail(self):
        with pytest.raises(exceptions.InfluxDBInvalidTypeError):
            Query()\
                .select()\
                .from_measurements('default')\
                .limit(-10)

    def test_with_bad_value_fail(self):
        with pytest.raises(exceptions.InfluxDBInvalidTypeError):
            Query()\
                .select()\
                .from_measurements('default')\
                .limit(True)

    def test_with_decimal_value_fail(self):
        with pytest.raises(exceptions.InfluxDBInvalidTypeError):
            Query()\
                .select()\
                .from_measurements('default')\
                .limit(10.5)


class TestDBSLimitQuery:
    def test_with_good_value_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .slimit(10)
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT * FROM "default" SLIMIT 10'
        res = query.execute()
        assert 'results' in res

    def test_chain_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .slimit(10)\
            .slimit(100)
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT * FROM "default" SLIMIT 100'
        res = query.execute()
        assert 'results' in res

    def test_with_empty_fail(self):
        with pytest.raises(TypeError):
            Query()\
                .select()\
                .from_measurements('default')\
                .slimit()

    def test_with_negative_value_fail(self):
        with pytest.raises(exceptions.InfluxDBInvalidTypeError):
            Query()\
                .select()\
                .from_measurements('default')\
                .slimit(-10)

    def test_with_bad_value_fail(self):
        with pytest.raises(exceptions.InfluxDBInvalidTypeError):
            Query()\
                .select()\
                .from_measurements('default')\
                .slimit(True)

    def test_with_decimal_value_fail(self):
        with pytest.raises(exceptions.InfluxDBInvalidTypeError):
            Query()\
                .select()\
                .from_measurements('default')\
                .slimit(10.5)


class TestDBOffsetQuery:
    def test_with_good_value_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .offset(10)
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT * FROM "default" OFFSET 10'
        res = query.execute()
        assert 'results' in res

    def test_chain_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .offset(10)\
            .offset(100)
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT * FROM "default" OFFSET 100'
        res = query.execute()
        assert 'results' in res

    def test_with_empty_fail(self):
        with pytest.raises(TypeError):
            Query()\
                .select()\
                .from_measurements('default')\
                .offset()

    def test_with_negative_value_fail(self):
        with pytest.raises(exceptions.InfluxDBInvalidTypeError):
            Query()\
                .select()\
                .from_measurements('default')\
                .offset(-10)

    def test_with_bad_value_fail(self):
        with pytest.raises(exceptions.InfluxDBInvalidTypeError):
            Query()\
                .select()\
                .from_measurements('default')\
                .offset(True)

    def test_with_decimal_value_fail(self):
        with pytest.raises(exceptions.InfluxDBInvalidTypeError):
            Query()\
                .select()\
                .from_measurements('default')\
                .offset(10.5)


class TestDBSOffsetQuery:
    def test_with_good_value_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .soffset(10)
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT * FROM "default" SOFFSET 10'
        res = query.execute()
        assert 'results' in res

    def test_chain_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .soffset(10)\
            .soffset(100)
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT * FROM "default" SOFFSET 100'
        res = query.execute()
        assert 'results' in res

    def test_with_empty_fail(self):
        with pytest.raises(TypeError):
            Query()\
                .select()\
                .from_measurements('default')\
                .soffset()

    def test_with_negative_value_fail(self):
        with pytest.raises(exceptions.InfluxDBInvalidTypeError):
            Query()\
                .select()\
                .from_measurements('default')\
                .soffset(-10)

    def test_with_bad_value_fail(self):
        with pytest.raises(exceptions.InfluxDBInvalidTypeError):
            Query()\
                .select()\
                .from_measurements('default')\
                .soffset(True)

    def test_with_decimal_value_fail(self):
        with pytest.raises(exceptions.InfluxDBInvalidTypeError):
            Query()\
                .select()\
                .from_measurements('default')\
                .soffset(10.5)


class TestGroupByTagQuery:
    def test_with_all_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .group_by()
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT * FROM "default" GROUP BY *'
        res = query.execute()
        assert 'results' in res

    def test_with_one_tag_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .group_by('tag_1')
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT * FROM "default" GROUP BY tag_1'
        res = query.execute()
        assert 'results' in res

    def test_with_two_tag_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .group_by('tag_1', 'tag_2')
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT * FROM "default" GROUP BY tag_1, tag_2'
        res = query.execute()
        assert 'results' in res

    def test_chain_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .group_by('tag_1', 'tag_2')\
            .group_by('tag_3')
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT * FROM "default" GROUP BY tag_3'
        res = query.execute()
        assert 'results' in res

    def test_with_bad_value_fail(self):
        with pytest.raises(exceptions.InfluxDBInvalidTypeError):
            Query()\
                .select()\
                .from_measurements('default')\
                .group_by(True)


class TestGroupByTimeQuery:
    def test_with_good_interval_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .range_by('12s')
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT * FROM "default" GROUP BY time(12s)'
        res = query.execute()
        assert 'results' in res

    def test_with_good_shift_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .range_by('12s', shift='1d')
        prepared_query = query._get_prepared_query()
        expected_query = 'SELECT * FROM "default" GROUP BY time(12s,1d)'
        assert prepared_query == expected_query
        res = query.execute()
        assert 'results' in res

    def test_with_negative_shift_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .range_by('12s', shift='-1d')
        prepared_query = query._get_prepared_query()
        expected_query = 'SELECT * FROM "default" GROUP BY time(12s,-1d)'
        assert prepared_query == expected_query
        res = query.execute()
        assert 'results' in res

    def test_with_one_tag_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .range_by('12s', tags=['tag1'])
        prepared_query = query._get_prepared_query()
        expected_query = 'SELECT * FROM "default" GROUP BY time(12s),tag1'
        assert prepared_query == expected_query
        res = query.execute()
        assert 'results' in res

    def test_with_two_tags_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .range_by('12s', tags=['tag1', 'tag2'])
        prepared_query = query._get_prepared_query()
        expected_query = 'SELECT * FROM "default" GROUP BY time(12s),tag1,tag2'
        assert prepared_query == expected_query
        res = query.execute()
        assert 'results' in res

    def test_with_fill_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .range_by('12s', fill=3)
        prepared_query = query._get_prepared_query()
        expected_query = 'SELECT * FROM "default" GROUP BY time(12s) fill(3)'
        assert prepared_query == expected_query
        res = query.execute()
        assert 'results' in res

    def test_with_shift_and_fill_and_tags_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .range_by('12s', shift='1d', tags=['tag1'], fill=3)
        prepared_query = query._get_prepared_query()
        expected_query = 'SELECT * FROM "default" GROUP BY time(12s,1d),tag1 fill(3)'
        assert prepared_query == expected_query
        res = query.execute()
        assert 'results' in res

    def test_chain_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .range_by('12s')\
            .range_by('5s')
        prepared_query = query._get_prepared_query()
        expected_query = 'SELECT * FROM "default" GROUP BY time(5s)'
        assert prepared_query == expected_query
        res = query.execute()
        assert 'results' in res

    def test_with_fill_fail(self):
        with pytest.raises(exceptions.InfluxDBInvalidTypeError):
            Query()\
                .select()\
                .from_measurements('default')\
                .range_by('12s', fill=True)

    def test_with_bad_tags_fail(self):
        with pytest.raises(exceptions.InfluxDBInvalidTypeError):
            Query()\
                .select()\
                .from_measurements('default')\
                .range_by('12s', tags=True)

    def test_with_bad_tags_fail_2(self):
        with pytest.raises(exceptions.InfluxDBInvalidTypeError):
            Query()\
                .select()\
                .from_measurements('default')\
                .range_by('12s', tags=[1, 2])

    def test_with_bad_interval(self):
        with pytest.raises(exceptions.InfluxDBBadQueryError):
            query = Query()\
                .select()\
                .from_measurements('default')\
                .range_by('12k')
            query.execute()

    def test_with_bad_shift(self):
        with pytest.raises(exceptions.InfluxDBBadQueryError):
            query = Query()\
                .select()\
                .from_measurements('default')\
                .range_by('12s', shift='12k')
            query.execute()



class TestOrderByQuery:
    def test_asc_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .asc()
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT * FROM "default" ORDER BY ASC'
        res = query.execute()
        assert 'results' in res

    def test_desc_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .desc()
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT * FROM "default" ORDER BY DESC'
        res = query.execute()
        assert 'results' in res

    def test_chain_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .desc()\
            .asc()\
            .desc()
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT * FROM "default" ORDER BY DESC'
        res = query.execute()
        assert 'results' in res


class TestTimezoneQuery:
    def test_with_good_value_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .tz('Europe/Paris')
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT * FROM "default" tz(\'Europe/Paris\')'
        res = query.execute()
        assert 'results' in res

    def test_chain_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .tz('Europe/London')\
            .tz('Europe/Paris')
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT * FROM "default" tz(\'Europe/Paris\')'
        res = query.execute()
        assert 'results' in res

    def test_with_bad_value_fail(self):
        with pytest.raises(exceptions.InfluxDBInvalidTypeError):
            Query()\
                .select()\
                .from_measurements('default')\
                .tz(True)

    def test_with_bad_timezone_fail(self):
        with pytest.raises(exceptions.InfluxDBInvalidTypeError):
            Query()\
                .select()\
                .from_measurements('default')\
                .tz('Paris')


class TestSelectAggregation:
    def test_count_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .count()
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT COUNT(*) FROM "default"'
        res = query.execute()
        assert 'results' in res

    def test_count_with_value_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .count('value')
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT COUNT(value) FROM "default"'
        res = query.execute()
        assert 'results' in res

    def test_distinct_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .distinct()
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT DISTINCT(*) FROM "default"'
        res = query.execute()
        assert 'results' in res

    def test_mean_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .mean()
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT MEAN(*) FROM "default"'
        res = query.execute()
        assert 'results' in res

    def test_sum_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .sum()
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT SUM(*) FROM "default"'
        res = query.execute()
        assert 'results' in res

    def test_median_success(self):
        query = Query()\
            .select()\
            .from_measurements('default')\
            .median()
        prepared_query = query._get_prepared_query()
        assert prepared_query == 'SELECT MEDIAN(*) FROM "default"'
        res = query.execute()
        assert 'results' in res



import json
import requests
from . import exceptions


def raise_if_error(func):
    def func_wrapper(*args, **kwargs):
        try:
            request = args[0]
            params = kwargs.get('params', {})
            res = func(*args, **kwargs)
            try:
                json_res = res.json()
            except json.decoder.JSONDecodeError:
                json_res = {}
            res.raise_for_status()

        except requests.exceptions.MissingSchema as err:
            raise exceptions.InfluxDBInvalidURLError(request.base_url)

        except requests.exceptions.ConnectionError as err:
            raise exceptions.InfluxDBConnectionError(err)

        except requests.exceptions.HTTPError as err:
            if json_res and 'error' in json_res and\
               json_res['error'].startswith('error parsing query:'):
                query = params['q']
                error = json_res['error'][len('error parsing query:'):]
                raise exceptions.InfluxDBBadQueryError(query, error)

            if json_res and 'error' in json_res and\
               json_res['error'].endswith('invalid number'):
                points = kwargs['data']
                raise exceptions.InfluxDBInvalidNumberError(points)

            if json_res and 'error' in json_res and\
               json_res['error'].endswith('bad timestamp'):
                points = kwargs['data']
                raise exceptions.InfluxDBInvalidTimestampError(points)

            if res.status_code == 400:
                query = params['q']
                if query == '':
                    raise exceptions.InfluxDBEmptyRequestError(params)
                raise exceptions.InfluxDBBadRequestError(params)
            if res.status_code == 401:
                raise exceptions.InfluxDBUnauthorizedError(err)
            raise err
        return res
    return func_wrapper



from functools import lru_cache
from .criteria import Criteria
from .function import aggregations
from ..response import InfluxDBResponse
from ..serializers import BaseSerializer
from .. import Influxable, exceptions


class RawQuery:
    def __init__(self, str_query=''):
        self.str_query = str_query

    def execute(self):
        return self.raw_response

    @property
    def query(self):
        return self.str_query

    @property
    def raw_response(self):
        return self._resolve()

    @lru_cache(maxsize=None)
    def _resolve(self, *args, **kwargs):
        instance = Influxable.get_instance()
        return instance.execute_query(query=self.str_query, method='post')


class SelectQueryClause:
    def __init__(self):
        super(SelectQueryClause, self).__init__()
        self.select_clause = 'SELECT {fields}'
        self.selected_fields = []
        self.default_selected_fields = '*'

    def validate_field(self, field):
        if not isinstance(field, str):
            msg = 'field type must be <str>'
            raise exceptions.InfluxDBInvalidTypeError(msg)

    def select(self, *fields):
        selected_fields = []
        for field in fields:
            if not hasattr(field, 'evaluate'):
                self.validate_field(field)
                evaluated_field = str(field)
            else:
                evaluated_field = field.evaluate()
            selected_fields.append(evaluated_field)
        self.selected_fields = selected_fields
        return self

    def _prepare_select_clause(self):
        if len(self.selected_fields):
            joined_fields = ','.join(self.selected_fields)
        else:
            joined_fields = self.default_selected_fields
        return self.select_clause.format(fields=joined_fields)


class IntoQueryClause:
    def __init__(self):
        super(IntoQueryClause, self).__init__()
        self.into_clause = 'INTO {measurement}'
        self.selected_into_measurement = None

    def validate_measurement(self, measurement):
        if not isinstance(measurement, str):
            msg = 'measurement type must be <str>'
            raise exceptions.InfluxDBInvalidTypeError(msg)

    def into(self, measurement):
        self.validate_measurement(measurement)
        self.selected_into_measurement = measurement
        return self

    def _prepare_into_clause(self):
        into_clause = ''
        if self.selected_into_measurement is not None:
            into_clause = self.into_clause.format(
                measurement=self.selected_into_measurement,
            )
        return into_clause


class FromQueryClause:
    def __init__(self):
        super(FromQueryClause, self).__init__()
        self.from_clause = 'FROM {measurements}'
        self.selected_measurements = 'default'

    def validate_measurements(self, measurements):
        if len(measurements) == 0:
            msg = 'measurements should not be empty'
            raise exceptions.InfluxDBInvalidTypeError(msg)
        for measurement in measurements:
            if not isinstance(measurement, str):
                msg = 'measurement type must be <str>'
                raise exceptions.InfluxDBInvalidTypeError(msg)

    def from_measurements(self, *measurements):
        self.validate_measurements(measurements)
        quoted_measurements = ['"{}"'.format(m) for m in measurements]
        self.selected_measurements = ','.join(quoted_measurements)
        return self

    def _prepare_from_clause(self):
        return self.from_clause.format(measurements=self.selected_measurements)


class WhereQueryClause:
    def __init__(self):
        super(WhereQueryClause, self).__init__()
        self.where_clause = 'WHERE {criteria}'
        self.selected_criteria = []

    def validate_criteria(self, criteria):
        if len(criteria) == 0:
            msg = 'criteria should not be empty'
            raise exceptions.InfluxDBInvalidTypeError(msg)
        for c in criteria:
            if not isinstance(c, Criteria):
                msg = 'Invalid criteria'
                raise exceptions.InfluxDBInvalidTypeError(msg)

    def where(self, *criteria):
        self.validate_criteria(criteria)
        self.selected_criteria = list(criteria)
        return self

    def _prepare_where_clause(self):
        where_clause = ''
        if len(self.selected_criteria):
            criteria = [c.evaluate() for c in self.selected_criteria]
            eval_criteria = ' AND '.join(criteria)
            where_clause = self.where_clause.format(criteria=eval_criteria)
        return where_clause


class LimitQueryClause:
    def __init__(self):
        super(LimitQueryClause, self).__init__()
        self.limit_value = None

    def validate_value(self, value):
        if type(value) != int or value <= 0:
            msg = 'value must be a positive integer'
            raise exceptions.InfluxDBInvalidTypeError(msg)

    def limit(self, value):
        self.validate_value(value)
        self.limit_value = value
        return self

    def _prepare_limit_clause(self):
        limit_clause = ''
        if self.limit_value is not None:
            limit_clause = 'LIMIT {}'.format(self.limit_value)
        return limit_clause


class SLimitQueryClause:
    def __init__(self):
        super(SLimitQueryClause, self).__init__()
        self.slimit_value = None

    def validate_value(self, value):
        if type(value) != int or value <= 0:
            msg = 'value must be a positive integer'
            raise exceptions.InfluxDBInvalidTypeError(msg)

    def slimit(self, value):
        self.validate_value(value)
        self.slimit_value = value
        return self

    def _prepare_slimit_clause(self):
        slimit_clause = ''
        if self.slimit_value is not None:
            slimit_clause = 'SLIMIT {}'.format(self.slimit_value)
        return slimit_clause


class OffsetQueryClause:
    def __init__(self):
        super(OffsetQueryClause, self).__init__()
        self.offset_value = None

    def validate_value(self, value):
        if type(value) != int or value <= 0:
            msg = 'value must be a positive integer'
            raise exceptions.InfluxDBInvalidTypeError(msg)

    def offset(self, value):
        self.validate_value(value)
        self.offset_value = value
        return self

    def _prepare_offset_clause(self):
        offset_clause = ''
        if self.offset_value is not None:
            offset_clause = 'OFFSET {}'.format(self.offset_value)
        return offset_clause


class SOffsetQueryClause:
    def __init__(self):
        super(SOffsetQueryClause, self).__init__()
        self.soffset_value = None

    def validate_value(self, value):
        if type(value) != int or value <= 0:
            msg = 'value must be a positive integer'
            raise exceptions.InfluxDBInvalidTypeError(msg)

    def soffset(self, value):
        self.validate_value(value)
        self.soffset_value = value
        return self

    def _prepare_soffset_clause(self):
        soffset_clause = ''
        if self.soffset_value is not None:
            soffset_clause = 'SOFFSET {}'.format(self.soffset_value)
        return soffset_clause


class GroupByQueryClause:
    def __init__(self):
        super(GroupByQueryClause, self).__init__()
        self.group_by_clause = 'GROUP BY {}'
        self.time_subclause = 'time({params})'
        self.fill_subclause = 'fill({fill})'
        self.selected_group_by_tags = ['*']
        self.has_group_by_tags = False
        self.has_group_by_time = False
        self.interval_value = None
        self.shift_value = None
        self.fill_value = None

    def validate_tags(self, tags):
        if len(tags) == 0:
            msg = 'tags should not be empty'
            raise exceptions.InfluxDBInvalidTypeError(msg)
        for tag in tags:
            if not isinstance(tag, str):
                msg = 'tag type must be <str>'
                raise exceptions.InfluxDBInvalidTypeError(msg)

    def validate_fill(self, value):
        if type(value) != int:
            msg = 'fill type must be numeric'
            raise exceptions.InfluxDBInvalidTypeError(msg)

    def validate_range_by(self, interval, shift, fill, tags):
        if not isinstance(tags, list) and not isinstance(tags, tuple):
            msg = 'tags must be a list'
            raise exceptions.InfluxDBInvalidTypeError(msg)
        if len(tags):
            self.validate_tags(tags)
        if fill is not None:
            self.validate_fill(fill)

    def group_by(self, *tags):
        if len(tags):
            self.validate_tags(tags)
            self.selected_group_by_tags = list(tags)
        self.has_group_by_time = False
        self.has_group_by_tags = True
        return self

    def range_by(self, interval, shift=None, fill=None, tags=[]):
        self.validate_range_by(interval, shift, fill, tags)
        if len(tags):
            self.selected_group_by_tags = tags
        else:
            self.selected_group_by_tags = []
        self.has_group_by_tags = False
        self.has_group_by_time = True
        self.interval_value = interval
        self.shift_value = shift
        self.fill_value = fill
        return self

    def _prepare_time_subclause(self):
        if self.shift_value:
            params = ','.join([self.interval_value, self.shift_value])
        else:
            params = self.interval_value
        time_subclause = self.time_subclause.format(params=params)
        return time_subclause

    def _prepare_fill_subclause(self):
        fill_subclause = ''
        if self.fill_value:
            fill_subclause = self.fill_subclause.format(fill=self.fill_value)
        return fill_subclause

    def _prepare_group_by_tags_clause(self):
        tags = ', '.join(self.selected_group_by_tags)
        group_by_clause = self.group_by_clause.format(tags)
        return group_by_clause

    def _prepare_group_by_time_clause(self):
        tags = ','.join(self.selected_group_by_tags)
        time_subclause = self._prepare_time_subclause()
        fill_subclause = self._prepare_fill_subclause()
        temp_clause = time_subclause
        if tags:
            temp_clause += ','
            temp_clause += tags
        if fill_subclause:
            temp_clause += ' '
            temp_clause += fill_subclause
        group_by_clause = self.group_by_clause.format(temp_clause)
        return group_by_clause

    def _prepare_group_by_clause(self):
        group_by_clause = ''
        if self.has_group_by_tags:
            group_by_clause = self._prepare_group_by_tags_clause()
        if self.has_group_by_time:
            group_by_clause = self._prepare_group_by_time_clause()
        return group_by_clause


class OrderByQueryClause:
    def __init__(self):
        super(OrderByQueryClause, self).__init__()
        self.is_chronological_sort = None

    def asc(self):
        self.is_chronological_sort = True
        return self

    def desc(self):
        self.is_chronological_sort = False
        return self

    def _prepare_order_by_clause(self):
        order_by_clause = ''
        if self.is_chronological_sort is not None:
            order_by_value = 'ASC' if self.is_chronological_sort else 'DESC'
            order_by_clause = 'ORDER BY {}'.format(order_by_value)
        return order_by_clause


class TimezoneClause:
    def __init__(self):
        super(TimezoneClause, self).__init__()
        self.timezone_clause = 'tz(\'{timezone}\')'
        self.timezone_value = None

    def validate_timezone(self, value):
        import pytz
        if not isinstance(value, str):
            msg = 'value must be a timezone string'
            raise exceptions.InfluxDBInvalidTypeError(msg)

        if value not in pytz.all_timezones:
            msg = 'value is an invalid timezone'
            raise exceptions.InfluxDBInvalidTypeError(msg)

    def tz(self, value):
        self.validate_timezone(value)
        self.timezone_value = value
        return self

    def _prepare_timezone_clause(self):
        timezone_value = ''
        if self.timezone_value is not None:
            timezone_value = self.timezone_clause.format(
                timezone=self.timezone_value,
            )
        return timezone_value


class SelectAggregation:
    def count(self, value='*'):
        return self.select(aggregations.Count(value))

    def distinct(self, value='*'):
        return self.select(aggregations.Distinct(value))

    def integral(self, value='*'):
        return self.select(aggregations.Integral(value))

    def mean(self, value='*'):
        return self.select(aggregations.Mean(value))

    def median(self, value='*'):
        return self.select(aggregations.Median(value))

    def mode(self, value='*'):
        return self.select(aggregations.Mode(value))

    def spread(self, value='*'):
        return self.select(aggregations.Spread(value))

    def std_dev(self, value='*'):
        return self.select(aggregations.StdDev(value))

    def sum(self, value='*'):
        return self.select(aggregations.Sum(value))


class GenericQuery(
    SelectQueryClause,
    IntoQueryClause,
    FromQueryClause,
    WhereQueryClause,
    LimitQueryClause,
    SLimitQueryClause,
    OffsetQueryClause,
    SOffsetQueryClause,
    GroupByQueryClause,
    OrderByQueryClause,
    TimezoneClause,
    SelectAggregation,
):
    def __init__(self):
        super(GenericQuery, self).__init__()


class Query(GenericQuery, RawQuery):
    def __init__(self):
        super(Query, self).__init__()

    def _get_initial_query(self):
        initial_query = ' '.join([
            '{select_clause}',
            '{into_clause}',
            '{from_clause}',
            '{where_clause}',
            '{limit_clause}',
            '{offset_clause}',
            '{slimit_clause}',
            '{soffset_clause}',
            '{group_by_clause}',
            '{order_by_clause}',
            '{timezone_clause}',
        ])
        return initial_query

    def _get_prepared_query(self):
        initial_query = self._get_initial_query()
        select_clause = self._prepare_select_clause()
        into_clause = self._prepare_into_clause()
        from_clause = self._prepare_from_clause()
        where_clause = self._prepare_where_clause()
        limit_clause = self._prepare_limit_clause()
        offset_clause = self._prepare_offset_clause()
        slimit_clause = self._prepare_slimit_clause()
        soffset_clause = self._prepare_soffset_clause()
        group_by_clause = self._prepare_group_by_clause()
        order_by_clause = self._prepare_order_by_clause()
        timezone_clause = self._prepare_timezone_clause()
        prepared_query = initial_query.format(
            select_clause=select_clause,
            into_clause=into_clause,
            from_clause=from_clause,
            where_clause=where_clause,
            limit_clause=limit_clause,
            offset_clause=offset_clause,
            slimit_clause=slimit_clause,
            soffset_clause=soffset_clause,
            group_by_clause=group_by_clause,
            order_by_clause=order_by_clause,
            timezone_clause=timezone_clause,
        )
        prepared_query = ' '.join(prepared_query.split())
        prepared_query = prepared_query.strip()
        return prepared_query

    def execute(self):
        prepared_query = self._get_prepared_query()
        # print('prepared_query', prepared_query)
        self.str_query = prepared_query
        return super().execute()

    def format(self, result, parser_class=BaseSerializer, **kwargs):
        return parser_class(result, **kwargs).convert()

    def evaluate(self, parser_class=BaseSerializer, **kwargs):
        result = InfluxDBResponse(self.execute())
        result.raise_if_error()
        formatted_result = self.format(result, parser_class, **kwargs)
        return formatted_result


class BulkInsertQuery(RawQuery):
    @lru_cache(maxsize=None)
    def _resolve(self, *args, **kwargs):
        instance = Influxable.get_instance()
        return instance.write_points(points=self.str_query)



import json
import pandas as pd
from influxable import attributes, serializers
from influxable.db import RawQuery
from influxable.measurement import Measurement
from influxable.response import InfluxDBResponse


class TestSerializer:
    def create_measurement_class(self):
        class MySampleMeasurement(Measurement):
            measurement_name = 'mysamplemeasurement'
            time = attributes.TimestampFieldAttribute(precision="s")
            value = attributes.IntegerFieldAttribute()
        measurement_cls = MySampleMeasurement
        return measurement_cls

    def execute_query(self, query):
        response = RawQuery(query).execute()
        influx_response = InfluxDBResponse(response)
        return influx_response

    def execute_sample_query(self):
        query = 'SHOW DATABASES'
        influx_response = self.execute_query(query)
        return influx_response

    def test_base_serializer_success(self):
        influx_response = self.execute_sample_query()
        serializer = serializers.BaseSerializer(influx_response)
        data = serializer.convert()
        assert isinstance(data, dict)
        assert 'results' in data

    def test_json_serializer_success(self):
        influx_response = self.execute_sample_query()
        serializer = serializers.JsonSerializer(influx_response)
        data = serializer.convert()
        json_data = json.loads(data)
        assert isinstance(data, str)
        assert json_data is not None

    def test_formatted_serie_serializer_success(self):
        influx_response = self.execute_sample_query()
        serializer = serializers.FormattedSerieSerializer(influx_response)
        data = serializer.convert()
        assert isinstance(data, list)
        assert len(data) == 1
        assert 'databases' in data[0]

    def test_flat_formatted_serie_serializer_success(self):
        influx_response = self.execute_sample_query()
        serializer = serializers.FlatFormattedSerieSerializer(influx_response)
        data = serializer.convert()
        assert isinstance(data, list)
        for d in data:
            assert isinstance(d, dict)

    def test_flat_simple_result_serializer_success(self):
        influx_response = self.execute_sample_query()
        serializer = serializers.FlatSimpleResultSerializer(influx_response)
        data = serializer.convert()
        assert isinstance(data, list)
        for d in data:
            assert isinstance(d, str)

    def test_pandas_serializer_success(self):
        influx_response = self.execute_sample_query()
        serializer = serializers.PandasSerializer(influx_response)
        data = serializer.convert()
        assert type(data) == pd.DataFrame

    def test_measurement_serializer_success(self):
        measurement_cls = self.create_measurement_class()
        points = measurement_cls\
            .get_query()\
            .limit(10)\
            .evaluate()

        assert isinstance(points, list)
        for p in points:
            assert isinstance(p, measurement_cls)



import arrow
import pytest
from datetime import datetime
from decimal import Decimal as D
from influxable import attributes, exceptions


class TestBaseAttribute:
    def test_clean_success(self):
        base_attr = attributes.BaseAttribute()
        base_attr._value = 5
        base_attr.clean(5)
        value = base_attr.get_internal_value()
        assert value == 5

    def test_clean_with_none_value_success(self):
        base_attr = attributes.BaseAttribute()
        base_attr.clean(None)
        value = base_attr.get_internal_value()
        assert value is None

    def test_clean_with_default_value_success(self):
        base_attr = attributes.BaseAttribute(default=10)
        base_attr.clean(None)
        value = base_attr.get_internal_value()
        assert value == 10

    def test_clone_success(self):
        base_attr = attributes.BaseAttribute()
        cloned_attr = base_attr.clone()
        assert base_attr != cloned_attr

    def test_clone_with_value_success(self):
        base_attr = attributes.BaseAttribute()
        base_attr._value = 5
        cloned_attr = base_attr.clone()
        value = base_attr.get_internal_value()
        cloned_value = cloned_attr.get_internal_value()
        assert base_attr != cloned_attr
        assert value == cloned_value

    def test_get_internal_value_success(self):
        base_attr = attributes.BaseAttribute()
        base_attr._value = 5
        assert base_attr._value == 5
        assert base_attr.value == 5
        assert base_attr.get_internal_value() == 5

    def test_get_prep_value_success(self):
        base_attr = attributes.BaseAttribute()
        base_attr._value = 5
        assert base_attr.get_prep_value() == base_attr.to_influx(5)

    def test_get_name_success(self):
        base_attr = attributes.BaseAttribute(name='my_field')
        assert base_attr.attribute_name == 'my_field'
        assert base_attr.name == 'my_field'

    def test_reset_success(self):
        base_attr = attributes.BaseAttribute()
        base_attr._value = 5
        base_attr.reset()
        assert base_attr.get_internal_value() is None

    def test_set_internal_value_success(self):
        base_attr = attributes.BaseAttribute()
        base_attr.set_internal_value(5)
        assert base_attr.get_internal_value() == 5
        base_attr.set_internal_value(True)
        assert base_attr.get_internal_value() is True
        base_attr.set_internal_value('hello')
        assert base_attr.get_internal_value() == 'hello'
        base_attr.set_internal_value(None)
        assert base_attr.get_internal_value() is None

    def test_validate_is_nullable_fail(self):
        with pytest.raises(exceptions.InfluxDBAttributeValueError):
            base_attr = attributes.BaseAttribute(is_nullable=False)
            base_attr.set_internal_value(None)

    def test_validate_is_nullable_success(self):
        base_attr = attributes.BaseAttribute(is_nullable=False)
        base_attr.set_internal_value(5)


class TestIntegerFieldAttribute:
    def test_to_python_success(self):
        attr = attributes.IntegerFieldAttribute()
        assert attr.to_python(5) == 5

    def test_to_influx_success(self):
        attr = attributes.IntegerFieldAttribute()
        assert attr.to_influx(5) == '5i'

    def test_validate_min_value_fail(self):
        with pytest.raises(exceptions.InfluxDBAttributeValueError):
            base_attr = attributes.IntegerFieldAttribute(min_value=5)
            base_attr.set_internal_value(4)

    def test_validate_min_value_success(self):
        base_attr = attributes.IntegerFieldAttribute(min_value=5)
        base_attr.set_internal_value(6)

    def test_validate_max_value_fail(self):
        with pytest.raises(exceptions.InfluxDBAttributeValueError):
            base_attr = attributes.IntegerFieldAttribute(max_value=5)
            base_attr.set_internal_value(6)

    def test_validate_max_value_success(self):
        base_attr = attributes.IntegerFieldAttribute(max_value=5)
        base_attr.set_internal_value(4)

    def test_validate_invalid_min_value_fail(self):
        with pytest.raises(exceptions.InfluxDBAttributeValueError):
            attributes.IntegerFieldAttribute(min_value='ok')

    def test_validate_invalid_max_value_fail(self):
        with pytest.raises(exceptions.InfluxDBAttributeValueError):
            attributes.IntegerFieldAttribute(max_value='ok')


class TestFloatFieldAttribute:
    def test_clean_success(self):
        attr = attributes.FloatFieldAttribute()
        attr.set_internal_value(5.2345)
        assert attr.get_internal_value() == D(5.2345)

    def test_clean_with_max_nb_decimals_success(self):
        attr = attributes.FloatFieldAttribute(max_nb_decimals=2)
        attr.clean(5.2345)
        assert attr.get_internal_value() == D(5.23).quantize(D('.01'))

    def test_to_python_success(self):
        attr = attributes.FloatFieldAttribute()
        assert attr.to_python(5.2504) == 5.2504

    def test_to_influx_success(self):
        attr = attributes.FloatFieldAttribute()
        assert attr.to_influx(5.2504) == '5.2504'

    def test_validate_invalid_type_fail(self):
        with pytest.raises(exceptions.InfluxDBAttributeValueError):
            attributes.FloatFieldAttribute(max_nb_decimals='ok')

    def test_validate_negative_max_nb_decimals_fail(self):
        with pytest.raises(exceptions.InfluxDBAttributeValueError):
            attributes.FloatFieldAttribute(max_nb_decimals=-5)


class TestStringFieldAttribute:
    def test_to_python_success(self):
        attr = attributes.StringFieldAttribute()
        assert attr.to_python(5) == '5'

    def test_to_influx_success(self):
        attr = attributes.StringFieldAttribute()
        assert attr.to_influx('5') == "\'5\'"

    def test_validate_invalid_choices_type_fail(self):
        with pytest.raises(exceptions.InfluxDBAttributeValueError):
            attributes.StringFieldAttribute(choices=5)

    def test_validate_invalid_choices_item_type_fail(self):
        with pytest.raises(exceptions.InfluxDBAttributeValueError):
            attributes.StringFieldAttribute(choices=[5, 2, 3])

    def test_validate_bad_choice_type_fail(self):
        with pytest.raises(exceptions.InfluxDBAttributeValueError):
            attr = attributes.StringFieldAttribute(choices=['first', 'last'])
            attr.set_internal_value('ok')

    def test_validate_invalid_max_length_type_fail(self):
        with pytest.raises(exceptions.InfluxDBAttributeValueError):
            attributes.StringFieldAttribute(max_length='test')

    def test_validate_invalid_negative_max_length_fail(self):
        with pytest.raises(exceptions.InfluxDBAttributeValueError):
            attributes.StringFieldAttribute(max_length=-7)

    def test_validate_bad_max_length_fail(self):
        with pytest.raises(exceptions.InfluxDBAttributeValueError):
            attr = attributes.StringFieldAttribute(max_length=5)
            attr.set_internal_value('test_value')


class TestBooleanFieldAttribute:
    def test_to_python_success(self):
        attr = attributes.BooleanFieldAttribute()
        assert attr.to_python(True) is True

    def test_to_influx_success(self):
        attr = attributes.BooleanFieldAttribute()
        assert attr.to_influx(True) == "true"


class TestTimestampFieldAttribute:
    def test_to_python_success(self):
        attr = attributes.TimestampFieldAttribute()
        attr.set_internal_value(1570209691)
        assert attr.to_python(1570209691) == D('1570209691000000000')

    def test_to_influx_success(self):
        attr = attributes.TimestampFieldAttribute()
        attr.set_internal_value(1570209691)
        assert attr.to_influx(1570209691000000000) == '1570209691000000000'

    def test_clean_success(self):
        attr = attributes.TimestampFieldAttribute()
        attr.set_internal_value(1570209691)
        assert attr.to_python(1570209691) == D('1570209691000000000')

    def test_clean_with_auto_now_success(self):
        attr = attributes.TimestampFieldAttribute(auto_now=True)
        attr.set_internal_value(None)
        assert attr.get_internal_value() is not None

    def test_clean_with_none_success(self):
        attr = attributes.TimestampFieldAttribute(auto_now=False)
        attr.set_internal_value(None)
        assert attr.get_internal_value() is None

    def test_convert_to_nanoseconds_success(self):
        attr = attributes.TimestampFieldAttribute()
        attr.set_internal_value(1570209691)
        assert attr.convert_to_nanoseconds(1570209691) == D('1570209691000000000')

    def test_convert_to_precision_success(self):
        attr = attributes.TimestampFieldAttribute()
        attr.set_internal_value(1570209691)
        assert attr.convert_to_precision(1570209691, 'ms') == D('1570209691000')

    def test_validate_fail(self):
        with pytest.raises(exceptions.InfluxDBAttributeValueError):
            attributes.TimestampFieldAttribute(precision='k')


class TestDateTimeFieldAttribute:
    def test_to_python_success(self):
        attr = attributes.DateTimeFieldAttribute()
        attr.set_internal_value('2019-10-01 10:11:05')
        dt = arrow.get('2019-10-01 10:11:05', attr.str_format).datetime
        assert attr.to_python('2019-10-01 10:11:05') == dt

    def test_to_python_with_datetime_success(self):
        attr = attributes.DateTimeFieldAttribute()
        dt = arrow.get('2019-10-01 10:11:05', attr.str_format).datetime
        attr.set_internal_value(dt)
        assert attr.to_python(dt) == dt

    def test_to_influx_success(self):
        attr = attributes.DateTimeFieldAttribute()
        attr.set_internal_value('2019-10-04 21:20:34')
        dt = attr.get_internal_value()
        assert attr.to_influx(dt) == '1570224034000000000'

    def test_get_internal_value_success(self):
        attr = attributes.DateTimeFieldAttribute()
        attr.set_internal_value('2019-10-01 10:11:05')
        assert attr.get_internal_value() == '2019-10-01 10:11:05'

    def test_clean_success(self):
        attr = attributes.DateTimeFieldAttribute()
        attr.set_internal_value('2019-10-04 21:20:34')
        dt = arrow.get('2019-10-04 21:20:34', attr.str_format).datetime
        assert attr.to_python('2019-10-04 21:20:34') == dt

    def test_clean_with_auto_now_success(self):
        attr = attributes.DateTimeFieldAttribute(auto_now=True)
        attr.set_internal_value(None)
        assert attr.get_internal_value() is not None
        assert isinstance(attr._value, datetime)

    def test_clean_with_none_success(self):
        attr = attributes.DateTimeFieldAttribute(auto_now=False)
        attr.set_internal_value(None)
        assert attr.get_internal_value() is None



import pytest
from influxable import Influxable, exceptions
from influxable.db import admin, InfluxDBAdmin, Field
from influxable.response import InfluxDBResponse


class TestGenericDBAdminCommand:
    def test_add_database_name_to_options_success(self):
        options = admin.GenericDBAdminCommand._add_database_name_to_options({})
        assert 'database_name' in options

    def test_execute_query_success(self):
        query = 'SHOW DATABASES'
        influx_response = admin.GenericDBAdminCommand._execute_query(query)
        assert isinstance(influx_response, InfluxDBResponse)
        assert influx_response.raw is not None
        assert influx_response.error is None

    def test_execute_query_failed(self):
        with pytest.raises(exceptions.InfluxDBError):
            query = 'SHOW MEASUREMENTS WHERE time < 200'
            admin.GenericDBAdminCommand._execute_query(query)

    def test_execute_query_with_parser_success(self):
        query = 'SHOW DATABASES'
        res = admin.GenericDBAdminCommand._execute_query_with_parser(query)
        assert isinstance(res, list)

    def test_get_formatted_privilege_success(self):
        privilege = admin.GenericDBAdminCommand._get_formatted_privilege('all')
        assert privilege == 'ALL'

    def test_get_formatted_privilege_success_2(self):
        privilege = admin.GenericDBAdminCommand._get_formatted_privilege(
            admin.Privileges.ALL
        )
        assert privilege == 'ALL'

    def test_get_formatted_privilege_failed(self):
        with pytest.raises(exceptions.InfluxDBInvalidChoiceError):
            admin.GenericDBAdminCommand._get_formatted_privilege('none')

    def test_get_formatted_user_name_success(self):
        name = admin.GenericDBAdminCommand._get_formatted_user_name('javid')
        assert name == '"javid"'

    def test_generate_from_clause_success(self):
        clause = admin.GenericDBAdminCommand._generate_from_clause(['param1'])
        assert clause == 'FROM "param1"'

    def test_generate_from_clause_with_two_measurement_success(self):
        clause = admin.GenericDBAdminCommand._generate_from_clause(
            ['param1', 'param2']
        )
        assert clause == 'FROM "param1", "param2"'

    def test_generate_from_clause_without_measurement_success(self):
        clause = admin.GenericDBAdminCommand._generate_from_clause([])
        assert clause == ''

    def test_generate_from_clause_failed(self):
        with pytest.raises(exceptions.InfluxDBInvalidTypeError):
            admin.GenericDBAdminCommand._generate_from_clause(True)

    def test_generate_where_clause_success(self):
        clause = admin.GenericDBAdminCommand._generate_where_clause(
            [Field('time') < 20]
        )
        assert clause == 'WHERE "time" < 20'

    def test_generate_where_clause_with_two_measurement_success(self):
        clause = admin.GenericDBAdminCommand._generate_where_clause(
            [
                Field('time') < 20,
                Field('phase') == 'MOON',
            ]
        )
        assert clause == 'WHERE "time" < 20 AND "phase" = \'MOON\''

    def test_generate_where_clause_without_measurement_success(self):
        clause = admin.GenericDBAdminCommand._generate_where_clause([])
        assert clause == ''

    def test_generate_where_clause_failed(self):
        with pytest.raises(exceptions.InfluxDBInvalidTypeError):
            admin.GenericDBAdminCommand._generate_where_clause(True)

    def test_generate_limit_clause_success(self):
        clause = admin.GenericDBAdminCommand._generate_limit_clause(10)
        assert clause == 'LIMIT 10'

    def test_generate_limit_clause_failed(self):
        with pytest.raises(exceptions.InfluxDBInvalidTypeError):
            admin.GenericDBAdminCommand._generate_limit_clause("test")

    def test_generate_offset_clause_success(self):
        clause = admin.GenericDBAdminCommand._generate_offset_clause(10)
        assert clause == 'OFFSET 10'

    def test_generate_offset_clause_failed(self):
        with pytest.raises(exceptions.InfluxDBInvalidTypeError):
            admin.GenericDBAdminCommand._generate_offset_clause("test")

    def test_generate_default_clause_success_1(self):
        clause = admin.GenericDBAdminCommand._generate_default_clause(True)
        assert clause == 'DEFAULT'

    def test_generate_default_clause_success_2(self):
        clause = admin.GenericDBAdminCommand._generate_default_clause(False)
        assert clause == ''

    def test_generate_duration_clause_success_1(self):
        clause = admin.GenericDBAdminCommand._generate_duration_clause('1d')
        assert clause == 'DURATION 1d'

    def test_generate_duration_clause_success_2(self):
        clause = admin.GenericDBAdminCommand._generate_duration_clause()
        assert clause == ''

    def test_generate_replication_clause_success_1(self):
        clause = admin.GenericDBAdminCommand._generate_replication_clause(4)
        assert clause == 'REPLICATION 4'

    def test_generate_replication_clause_success_2(self):
        clause = admin.GenericDBAdminCommand._generate_replication_clause()
        assert clause == ''

    def test_generate_shard_duration_clause_success_1(self):
        clause = admin.GenericDBAdminCommand._generate_shard_duration_clause('4d')
        assert clause == 'SHARD DURATION 4d'

    def test_generate_shard_duration_clause_success_2(self):
        clause = admin.GenericDBAdminCommand._generate_shard_duration_clause()
        assert clause == ''

    def test_get_influxable_instance_success(self):
        instance = admin.GenericDBAdminCommand._get_influxable_instance()
        assert isinstance(instance, Influxable)

    def test_format_with_simple_quote_success(self):
        string = admin.GenericDBAdminCommand._format_with_simple_quote('test')
        assert string == '\'test\''

    def test_format_with_double_quote_success(self):
        string = admin.GenericDBAdminCommand._format_with_double_quote('test')
        assert string == '"test"'


class TestAlterAdminCommand:
    @classmethod
    def setup_class(cls):
        InfluxDBAdmin.create_retention_policy(
            'test_policy',
            duration='1d',
            replication=3,
        )

    def test_alter_retention_policy_success_1(self):
        res = InfluxDBAdmin.alter_retention_policy(
            'test_policy',
            duration='1d',
        )
        assert res is True

    def test_alter_retention_policy_success_2(self):
        res = InfluxDBAdmin.alter_retention_policy(
            'test_policy',
            replication=4,
        )
        assert res is True

    def test_alter_retention_policy_success_3(self):
        res = InfluxDBAdmin.alter_retention_policy(
            'test_policy',
            shard_duration='1h',
        )
        assert res is True

    def test_alter_retention_policy_failed(self):
        with pytest.raises(exceptions.InfluxDBError):
            InfluxDBAdmin.alter_retention_policy(
                'test_policy'
            )


class TestCreateAdminCommand:
    def test_create_database_success_1(self):
        res = InfluxDBAdmin.create_database(
            'test_database_1',
        )
        assert res is True

    def test_create_database_success_2(self):
        res = InfluxDBAdmin.create_database(
            'test_database_2',
            duration='1d',
        )
        assert res is True

    def test_create_database_success_3(self):
        res = InfluxDBAdmin.create_database(
            'test_database_3',
            replication=4,
        )
        assert res is True

    def test_create_database_success_4(self):
        res = InfluxDBAdmin.create_database(
            'test_database_4',
            shard_duration='1h',
        )
        assert res is True

    def test_create_database_success_5(self):
        res = InfluxDBAdmin.create_database(
            'test_database_5',
            policy_name='test_policy_2',
        )
        assert res is True

    def test_create_retention_policy_success_1(self):
        res = InfluxDBAdmin.create_retention_policy(
            'test_policy_1',
            duration='1d',
            replication=3,
        )
        assert res is True

    def test_create_retention_policy_success_2(self):
        res = InfluxDBAdmin.create_retention_policy(
            'test_policy_22',
            duration='1d',
            replication=4,
        )
        assert res is True

    def test_create_retention_policy_success_3(self):
        res = InfluxDBAdmin.create_retention_policy(
            'test_policy_3',
            duration='1d',
            replication=4,
            shard_duration='1h',
        )
        assert res is True

    def test_create_retention_policy_failed_1(self):
        with pytest.raises(exceptions.InfluxDBError):
            InfluxDBAdmin.create_retention_policy(
                'test_policy'
            )

    def test_create_retention_policy_failed_2(self):
        with pytest.raises(exceptions.InfluxDBError):
            InfluxDBAdmin.create_retention_policy(
                'test_policy',
                replication=4,
                shard_duration='1h',
            )

    def test_create_retention_policy_failed_3(self):
        with pytest.raises(exceptions.InfluxDBError):
            InfluxDBAdmin.create_retention_policy(
                'test_policy',
                duration='1d',
                shard_duration='1h',
            )

    def test_create_retention_policy_failed_4(self):
        with pytest.raises(exceptions.InfluxDBError):
            InfluxDBAdmin.create_retention_policy(
                'test_policy',
                replication=4,
            )

    def test_create_subscription_success_1(self):
        res = InfluxDBAdmin.create_subscription(
            'test_subscription_1',
            hosts=["http://localhost:5000"],
        )
        assert res is True
        InfluxDBAdmin.drop_subscription('test_subscription_1')

    def test_create_subscription_success_2(self):
        res = InfluxDBAdmin.create_subscription(
            'test_subscription_2',
            hosts=["http://localhost:5000", "http://localhost:6000"],
            any=True,
        )
        assert res is True
        InfluxDBAdmin.drop_subscription('test_subscription_2')

    def test_create_user_success_1(self):
        res = InfluxDBAdmin.create_user(
            user_name='user_test_1',
            password='password_test_1',
        )
        assert res is True

    def test_create_user_success_2(self):
        res = InfluxDBAdmin.create_user(
            user_name='user_test_2',
            password='password_test_2',
            with_privileges=True,
        )
        assert res is True


class TestDeleteAdminCommand:
    def test_delete_success_1(self):
        res = InfluxDBAdmin.delete(
            measurements=['mymeas'],
        )
        assert res is True

    def test_delete_success_2(self):
        res = InfluxDBAdmin.delete(
            criteria=[Field('time') < 10],
        )
        assert res is True

    def test_delete_failed(self):
        with pytest.raises(exceptions.InfluxDBError):
            InfluxDBAdmin.delete()


class TestDropAdminCommand:
    def test_drop_continuous_query_success(self):
        res = InfluxDBAdmin.drop_continuous_query('test_continuous_query')
        assert res is True

    def test_drop_database_success(self):
        res = InfluxDBAdmin.drop_database('test_database')
        assert res is True

    def test_drop_measurement_success(self):
        res = InfluxDBAdmin.drop_measurement('test_measurement')
        assert res is True

    def test_drop_retention_policy_success(self):
        res = InfluxDBAdmin.drop_retention_policy('test_policy')
        assert res is True

    def test_drop_series_success_1(self):
        res = InfluxDBAdmin.drop_series(measurements=['mymeas'])
        assert res is True

    def test_drop_series_success_2(self):
        res = InfluxDBAdmin.drop_series(criteria=[Field('phase') == 'moon'])
        assert res is True

    def test_drop_series_failed(self):
        with pytest.raises(exceptions.InfluxDBError):
            InfluxDBAdmin.drop_series()

    def test_drop_shard_success(self):
        res = InfluxDBAdmin.drop_shard(1)
        assert res is True

    def test_drop_subscription_success(self):
        pytest.skip("SKIP")
        res = InfluxDBAdmin.drop_subscription('test_subscription')
        assert res is True

    def test_drop_user_success(self):
        pytest.skip("SKIP")
        res = InfluxDBAdmin.drop_user('test_user')
        assert res is True


class TestExplainAdminCommand:
    def test_explain_success_1(self):
        query = 'SELECT * FROM mymeas LIMIT 10'
        res = InfluxDBAdmin.explain(query)
        assert res is not None

    def test_explain_success_2(self):
        query = 'SELECT * FROM mymeas LIMIT 10'
        res = InfluxDBAdmin.explain(query, analyze=True)
        assert res is not None


class TestGrantAdminCommand:
    def test_grant_success(self):
        res = InfluxDBAdmin.grant(admin.Privileges.ALL, 'user_test_1')
        assert res is not None


class TestRevokeAdminCommand:
    def test_revoke_success(self):
        res = InfluxDBAdmin.revoke(admin.Privileges.ALL, 'user_test_1')
        assert res is not None


class TestShowAdminCommand:
    def test_show_field_key_cardinality_success(self):
        res = InfluxDBAdmin.show_field_key_cardinality()
        assert res is not None

    def test_show_field_key_cardinality_with_exact_success(self):
        res = InfluxDBAdmin.show_field_key_cardinality(exact=True)
        assert res is not None

    def test_show_measurement_cardinality_success(self):
        res = InfluxDBAdmin.show_measurement_cardinality()
        assert res is not None

    def test_show_measurement_cardinality_with_exact_success(self):
        res = InfluxDBAdmin.show_measurement_cardinality(exact=True)
        assert res is not None

    def test_show_series_cardinality_success(self):
        res = InfluxDBAdmin.show_series_cardinality()
        assert res is not None

    def test_show_series_cardinality_with_exact_success(self):
        res = InfluxDBAdmin.show_series_cardinality(exact=True)
        assert res is not None

    def test_show_tag_key_cardinality_success(self):
        res = InfluxDBAdmin.show_tag_key_cardinality()
        assert res is not None

    def test_show_tag_key_cardinality_with_exact_success(self):
        res = InfluxDBAdmin.show_tag_key_cardinality(exact=True)
        assert res is not None

    def test_show_tag_values_cardinality_success(self):
        res = InfluxDBAdmin.show_tag_values_cardinality('phase')
        assert res is not None

    def test_show_tag_values_cardinality_with_exact_success(self):
        res = InfluxDBAdmin.show_tag_values_cardinality('phase', exact=True)
        assert res is not None

    def test_show_continuous_queries_success(self):
        res = InfluxDBAdmin.show_continuous_queries()
        assert res is not None

    def test_show_diagnostics_success(self):
        res = InfluxDBAdmin.show_diagnostics()
        assert res is not None

    def test_show_field_keys_success(self):
        res = InfluxDBAdmin.show_field_keys()
        assert res is not None

    def test_show_field_keys_with_measurement_success(self):
        res = InfluxDBAdmin.show_field_keys(
            measurements=['param1'],
        )
        assert res is not None

    def test_show_field_keys_with_multiple_measurement_success(self):
        res = InfluxDBAdmin.show_field_keys(
            measurements=['param1', 'param2'],
        )
        assert res is not None

    def test_show_grants_success(self):
        res = InfluxDBAdmin.show_grants('admin')
        assert res is not None

    def test_show_databases_success(self):
        res = InfluxDBAdmin.show_databases()
        assert res is not None

    def test_show_measurements_success(self):
        res = InfluxDBAdmin.show_measurements()
        assert res is not None

    def test_show_measurements_with_criteria_success(self):
        res = InfluxDBAdmin.show_measurements(
            criteria=[Field('phase') == 'moon'],
        )
        assert res is not None

    def test_show_measurements_with_criteria_success_2(self):
        res = InfluxDBAdmin.show_measurements(
            criteria=[
                Field('phase') == 'moon',
                Field('phase') == 'sun',
            ],
        )
        assert res is not None

    def test_show_queries_success(self):
        res = InfluxDBAdmin.show_queries()
        assert res is not None

    def test_show_retention_policies_success(self):
        res = InfluxDBAdmin.show_retention_policies()
        assert res is not None

    def test_show_series_success(self):
        res = InfluxDBAdmin.show_series()
        assert res is not None

    def test_show_series_with_measurements_success(self):
        res = InfluxDBAdmin.show_series(
            measurements=['param1', 'param2'],
        )
        assert res is not None

    def test_show_series_with_criteria_success(self):
        res = InfluxDBAdmin.show_series(
            criteria=[Field('phase') == 'moon'],
        )
        assert res is not None

    def test_show_series_with_limit_success(self):
        res = InfluxDBAdmin.show_series(
            limit=10,
        )
        assert res is not None

    def test_show_series_with_offset_success(self):
        res = InfluxDBAdmin.show_series(
            offset=10,
        )
        assert res is not None

    def test_show_series_with_args_success(self):
        res = InfluxDBAdmin.show_series(
            measurements=['param1', 'param2'],
            criteria=[Field('phase') == 'moon'],
            limit=10,
            offset=10,
        )
        assert res is not None

    def test_show_stats_success(self):
        res = InfluxDBAdmin.show_stats()
        assert res is not None

    def test_show_shards_success(self):
        res = InfluxDBAdmin.show_shards()
        assert res is not None

    def test_show_shard_groups_success(self):
        res = InfluxDBAdmin.show_shard_groups()
        assert res is not None

    def test_show_subscriptions_success(self):
        res = InfluxDBAdmin.show_subscriptions()
        assert res is not None

    def test_show_tag_keys_success(self):
        res = InfluxDBAdmin.show_tag_keys()
        assert res is not None

    def test_show_tag_keys_with_measurement_success(self):
        res = InfluxDBAdmin.show_tag_keys(measurements=['param1'])
        assert res is not None

    def test_show_tag_keys_with_measurements_success(self):
        res = InfluxDBAdmin.show_tag_keys(measurements=['param1', 'param2'])
        assert res is not None

    def test_show_tag_values_success(self):
        res = InfluxDBAdmin.show_tag_values('phase')
        assert res is not None

    def test_show_tag_values_with_measurement_success(self):
        res = InfluxDBAdmin.show_tag_values(
            'phase',
            measurements=['param1'],
        )
        assert res is not None

    def test_show_tag_values_with_measurements_success(self):
        res = InfluxDBAdmin.show_tag_values(
            'phase',
            measurements=['param1', 'param2'],
        )
        assert res is not None

    def test_show_users_success(self):
        res = InfluxDBAdmin.show_users()
        assert res is not None



from influxable import settings


class TestSettings:
    def check_if_variable_exist(self, variable_name):
        return hasattr(settings, variable_name)

    def test_check_url_exist(self):
        assert self.check_if_variable_exist('INFLUXDB_URL')

    def test_check_user_exist(self):
        assert self.check_if_variable_exist('INFLUXDB_USER')

    def test_check_password_exist(self):
        assert self.check_if_variable_exist('INFLUXDB_PASSWORD')

    def test_check_database_name_exist(self):
        assert self.check_if_variable_exist('INFLUXDB_DATABASE_NAME')



from .api import InfluxDBApi
from .connection import Connection
from .helpers.decorators import Singleton


@Singleton
class Influxable:
    def __init__(self, *args, **kwargs):
        self.connection = Connection(*args, **kwargs)

    def create_connection(self, *args, **kwargs):
        return Connection.create_connection(*args, **kwargs)

    def ping(self, *args, **kwargs):
        request = self.connection.request
        return InfluxDBApi.ping(request, *args, **kwargs)

    def execute_query(self, *args, **kwargs):
        request = self.connection.request
        return InfluxDBApi.execute_query(request, *args, **kwargs)

    def write_points(self, *args, **kwargs):
        request = self.connection.request
        return InfluxDBApi.write_points(request, *args, **kwargs)

    @property
    def base_url(self):
        return self.connection.base_url

    @property
    def database_name(self):
        return self.connection.database_name

    @property
    def full_database_name(self):
        return self.connection.full_database_name

    @property
    def policy_name(self):
        return self.connection.policy_name



import arrow
import click
from influxable.commands.auto_generate import AutoGenerateMeasurement
from influxable.commands.populate import Populate


DEFAULT_OUTPUT_FILE_NAME = 'auto_generate_measurement.py'

DEFAULT_INTERVAL_DELAY = 30
DEFAULT_NB_YEARS_RANGE = 5
DEFAULT_POPULATE_MIN_VALUE = 10
DEFAULT_POPULATE_MAX_VALUE = 30
DEFAULT_MEASUREMENT_NAME = 'populate_temperature'
DEFAULT_TAGS = ['moon', 'sun']


def get_default_start_date():
    end_date = arrow.now('Europe/Paris').replace(hour=0, minute=0, second=0)
    start_date = end_date.shift(years=-DEFAULT_NB_YEARS_RANGE)
    return start_date.format('YYYY-MM-DDTHH:mm:ss')


def get_default_end_date():
    end_date = arrow.now('Europe/Paris').replace(hour=0, minute=0, second=0)
    return end_date.format('YYYY-MM-DDTHH:mm:ss')


@click.group()
def main(*args, **kwargs):
    pass


@main.command(name='autogenerate')
@click.option(
    '-o',
    '--output',
    'output_file_name',
    default=DEFAULT_OUTPUT_FILE_NAME,
    help='name of the output file',
)
def auto_generate(*args, **kwargs):
    """
    This command will automatically generate measurement classes.
    """
    AutoGenerateMeasurement.run(*args, **kwargs)


@main.command(name='populate')
@click.option(
    '--min_value',
    'min_value',
    type=int,
    default=DEFAULT_POPULATE_MIN_VALUE,
    help='minimum value for the populate range',
)
@click.option(
    '--max_value',
    'max_value',
    type=int,
    default=DEFAULT_POPULATE_MAX_VALUE,
    help='maximum value for the populate range',
)
@click.option(
    '-s',
    '--start_date',
    'start_date',
    type=click.DateTime(),
    default=get_default_start_date(),
    help='the start date range of populate values',
)
@click.option(
    '-e',
    '--end_date',
    'end_date',
    type=click.DateTime(),
    default=get_default_end_date(),
    help='the end date range of populate values',
)
@click.option(
    '-m',
    '--max_count_of_values',
    'max_count_of_values',
    type=int,
    default=None,
    help='maximum limit of added values count',
)
@click.option(
    '-id',
    '--interval_delay',
    'interval_delay',
    type=int,
    default=DEFAULT_INTERVAL_DELAY,
    help='maximum value for the populate range ',
)
@click.option(
    '-t',
    '--tags',
    'tags',
    multiple=True,
    default=DEFAULT_TAGS,
    help='list of tags for the populate',
)
@click.option(
    '-mn',
    '--measurement_name',
    'measurement_name',
    default=DEFAULT_MEASUREMENT_NAME,
    help='name of the populate measurement',
)
def populate(*args, **kwargs):
    """
    This command will exectute the populate command.
    """
    Populate.run(*args, **kwargs)



class InfluxDBException(Exception):
    pass


class InfluxDBError(InfluxDBException):
    pass


class InfluxDBConnectionError(InfluxDBError):
    pass


class InfluxDBInvalidResponseError(InfluxDBError):
    pass


class InfluxDBInvalidChoiceError(InfluxDBError):
    pass


class InfluxDBInvalidTypeError(InfluxDBError):
    pass


class InfluxDBInvalidURLError(InfluxDBError):
    MESSAGE_PLACEHOLDER = 'Bad schema for : {base_url}'

    def __init__(self, base_url):
        self.message = self.MESSAGE_PLACEHOLDER.format(base_url=base_url)
        super().__init__(self.message)


class InfluxDBBadRequestError(InfluxDBError):
    MESSAGE_PLACEHOLDER = 'Bad request for params : {params}'

    def __init__(self, params):
        self.message = self.MESSAGE_PLACEHOLDER.format(params=params)
        super().__init__(self.message)


class InfluxDBEmptyRequestError(InfluxDBError):
    MESSAGE_PLACEHOLDER = 'Empty request'

    def __init__(self, params):
        self.message = self.MESSAGE_PLACEHOLDER
        super().__init__(self.message)


class InfluxDBBadQueryError(InfluxDBError):
    MESSAGE_PLACEHOLDER = 'Invalid query : {query} - {error}'

    def __init__(self, query, error):
        self.message = self.MESSAGE_PLACEHOLDER.format(
            query=query,
            error=error,
        )
        super().__init__(self.message)


class InfluxDBInvalidNumberError(InfluxDBError):
    MESSAGE_PLACEHOLDER = 'Invalid number : {points}'

    def __init__(self, points):
        self.message = self.MESSAGE_PLACEHOLDER.format(points=points)
        super().__init__(self.message)


class InfluxDBInvalidTimestampError(InfluxDBError):
    MESSAGE_PLACEHOLDER = 'Invalid timestamp : {points}'

    def __init__(self, points):
        self.message = self.MESSAGE_PLACEHOLDER.format(points=points)
        super().__init__(self.message)


class InfluxDBUnauthorizedError(InfluxDBError):
    MESSAGE = 'Authorization Failed (Bad credentials)'

    def __init__(self, message):
        self.message = self.MESSAGE
        super().__init__(self.MESSAGE)


class InfluxDBAttributeValueError(InfluxDBError):
    pass



import pytest
from influxable import Influxable


class TestApp:
    def create_instance_of_influxable(self):
        instance = Influxable()
        return instance

    def get_instance(self):
        instance = Influxable.get_instance()
        return instance

    def test_singleton_first_instance_success(self):
        first_instance = self.create_instance_of_influxable()
        second_instance = self.get_instance()
        assert first_instance == second_instance

    def test_singleton_get_instance_success(self):
        instance = self.get_instance()
        assert instance is not None

    def test_singleton_create_second_instance_error(self):
        with pytest.raises(TypeError):
            self.create_instance_of_influxable()

    def test_ping_success(self):
        instance = self.get_instance()
        res = instance.ping()
        assert res is True

    def test_execute_query_success(self):
        query = 'SHOW DATABASES'
        instance = self.get_instance()
        res = instance.execute_query(query=query)
        assert res != {}

    def test_write_points_success(self):
        instance = self.get_instance()
        points = 'mysensor,esn=12,phase=moon value=10 1463289075000000000'
        res = instance.write_points(points)
        assert res is True



from . import settings
from .api import InfluxDBApi
from .request import InfluxDBRequest


class Connection:
    def __init__(self, *args, **kwargs):
        self.base_url = kwargs.get('base_url', settings.INFLUXDB_URL)
        self.user = kwargs.get('user', settings.INFLUXDB_USER)
        self.password = kwargs.get('password', settings.INFLUXDB_PASSWORD)
        self.database_name = kwargs.get(
            'database_name',
            settings.INFLUXDB_DATABASE_NAME,
        )

        self.auth = (self.user, self.password)
        self.request = InfluxDBRequest(
            self.base_url,
            self.database_name,
            auth=self.auth,
        )
        self.stream = False
        self.check_if_connection_reached()

    @staticmethod
    def create(base_url, database_name, user='', password=''):
        return Connection(base_url, database_name, user, password)

    def check_if_connection_reached(self):
        query = 'SHOW DATABASES'
        InfluxDBApi.execute_query(self.request, query)

    @property
    def policy_name(self):
        return 'autogen'

    @property
    def full_database_name(self):
        return '"{}"."{}"'.format(self.database_name, self.policy_name)



from .. import Influxable, exceptions, serializers
from .criteria import Criteria
from .query import RawQuery
from ..response import InfluxDBResponse


class Privileges:
    ALL = 'ALL'
    READ = 'READ'
    WRITE = 'WRITE'


PRIVILEGE_VALUES = ['ALL', 'READ', 'WRITE']


class GenericDBAdminCommand:
    @staticmethod
    def _add_database_name_to_options(options):
        database_name = GenericDBAdminCommand._get_database_name()
        full_database_name = GenericDBAdminCommand._get_full_database_name()
        database_name = GenericDBAdminCommand._format_with_double_quote(
            database_name
        )
        options.update({
            'database_name': database_name,
            'full_database_name': full_database_name,
        })
        return options

    @staticmethod
    def _execute_query(query, options={}):
        options = GenericDBAdminCommand._add_database_name_to_options(options)
        prepared_query = query.format(**options)
        response = RawQuery(prepared_query).execute()
        influx_response = InfluxDBResponse(response)
        influx_response.raise_if_error()
        return influx_response

    @staticmethod
    def _execute_query_with_parser(
        query,
        parser=serializers.FlatFormattedSerieSerializer,
        options={},
    ):
        influx_response = GenericDBAdminCommand._execute_query(query, options)
        formatted_result = parser(influx_response).convert()
        return formatted_result

    @staticmethod
    def _get_formatted_privilege(privilege):
        privilege = str(privilege).upper()
        if privilege not in PRIVILEGE_VALUES:
            msg = 'privilege `{}` must be one of value of {}'.format(
                privilege,
                PRIVILEGE_VALUES,
            )
            raise exceptions.InfluxDBInvalidChoiceError(msg)
        return privilege

    @staticmethod
    def _get_formatted_user_name(user_name):
        return GenericDBAdminCommand._format_with_double_quote(user_name)

    @staticmethod
    def _generate_from_clause(measurements):
        if not isinstance(measurements, list):
            msg = 'measurements type must be <list>'
            raise exceptions.InfluxDBInvalidTypeError(msg)

        quoted_measurements = ['"{}"'.format(m) for m in measurements]
        selected_measurements = ', '.join(quoted_measurements)

        from_clause = ''
        if selected_measurements:
            from_clause = 'FROM {}'.format(selected_measurements)
        return from_clause

    @staticmethod
    def _generate_where_clause(criteria):
        if not isinstance(criteria, list):
            msg = 'criteria type must be <list>'
            raise exceptions.InfluxDBInvalidTypeError(msg)

        if len(criteria) and \
           not any([isinstance(c, Criteria) for c in criteria]):
            msg = 'criteria type must be <list> of <Criteria>'
            raise exceptions.InfluxDBInvalidTypeError(msg)

        where_clause = ''
        if len(criteria):
            selected_criteria = [c.evaluate() for c in criteria]
            eval_criteria = ' AND '.join(selected_criteria)
            where_clause = 'WHERE {}'.format(eval_criteria)
        return where_clause

    @staticmethod
    def _generate_limit_clause(limit):
        if limit is not None and not isinstance(limit, int):
            msg = 'limit type must be <int>'
            raise exceptions.InfluxDBInvalidTypeError(msg)

        limit_clause = ''
        if limit is not None:
            limit_clause = 'LIMIT {}'.format(limit)
        return limit_clause

    @staticmethod
    def _generate_offset_clause(offset):
        if offset is not None and not isinstance(offset, int):
            msg = 'offset type must be <int>'
            raise exceptions.InfluxDBInvalidTypeError(msg)

        offset_clause = ''
        if offset is not None:
            offset_clause = 'OFFSET {}'.format(offset)
        return offset_clause

    @staticmethod
    def _generate_default_clause(is_default=False):
        return 'DEFAULT' if is_default is True else ''

    @staticmethod
    def _generate_duration_clause(duration=None):
        return 'DURATION {}'.format(duration) if duration else ''

    @staticmethod
    def _generate_replication_clause(replication=None):
        return 'REPLICATION {}'.format(replication) if replication else ''

    @staticmethod
    def _generate_shard_duration_clause(sh_duration=None):
        return 'SHARD DURATION {}'.format(sh_duration) if sh_duration else ''

    @staticmethod
    def _get_database_name():
        instance = GenericDBAdminCommand._get_influxable_instance()
        database_name = instance.database_name
        return database_name

    @staticmethod
    def _get_full_database_name():
        instance = GenericDBAdminCommand._get_influxable_instance()
        full_database_name = instance.full_database_name
        return full_database_name

    @staticmethod
    def _get_influxable_instance():
        instance = Influxable.get_instance()
        return instance

    @staticmethod
    def _format_with_simple_quote(string):
        return '\'{}\''.format(string)

    @staticmethod
    def _format_with_double_quote(string):
        return '"{}"'.format(string)


class AlterAdminCommand:
    @staticmethod
    def alter_retention_policy(
        policy_name,
        duration=None,
        replication=None,
        shard_duration=None,
        is_default=False,
    ):
        if not duration and not replication and not shard_duration\
           and not is_default:
            msg = '`duration` or `replication` or `shard_duration` ' +\
                  ' or `is_default` must be not null'
            raise exceptions.InfluxDBError(msg)

        policy_name = GenericDBAdminCommand._format_with_double_quote(
            policy_name,
        )

        default_clause = GenericDBAdminCommand._generate_default_clause(
            is_default
        )
        duration_clause = GenericDBAdminCommand._generate_duration_clause(
            duration
        )
        replication_clause = GenericDBAdminCommand._generate_replication_clause(
            replication
        )
        shard_duration_clause = GenericDBAdminCommand._generate_shard_duration_clause(
            shard_duration
        )

        options = {
            'default_clause': default_clause,
            'duration_clause': duration_clause,
            'policy_name': policy_name,
            'replication_clause': replication_clause,
            'shard_duration_clause': shard_duration_clause,
        }
        query = 'ALTER RETENTION POLICY {policy_name} ON {database_name}' +\
                ' {duration_clause}' +\
                ' {replication_clause}' +\
                ' {shard_duration_clause}' +\
                ' {default_clause}'
        InfluxDBAdmin._execute_query(query, options)
        return True


class CreateAdminCommand:
    @staticmethod
    def create_continuous_query():
        raise NotImplementedError

    @staticmethod
    def create_database(
        new_database_name,
        duration=None,
        replication=None,
        shard_duration=None,
        policy_name=None,
    ):
        new_database_name = GenericDBAdminCommand._format_with_double_quote(
            new_database_name,
        )
        options = {'new_database_name': new_database_name}

        with_clause = ''
        if duration or replication or shard_duration or policy_name:
            policy_clause = ''
            if policy_name:
                policy_clause = 'NAME "{}"'.format(policy_name)

            duration_clause = GenericDBAdminCommand._generate_duration_clause(
                duration
            )
            replication_clause = GenericDBAdminCommand._generate_replication_clause(
                replication
            )
            shard_duration_clause = GenericDBAdminCommand._generate_shard_duration_clause(
                shard_duration
            )

            with_clause = 'WITH' +\
                          ' {duration_clause}' +\
                          ' {replication_clause}' +\
                          ' {shard_duration_clause}' +\
                          ' {policy_clause}'
            options.update({
                'duration_clause': duration_clause,
                'replication_clause': replication_clause,
                'policy_clause': policy_clause,
                'shard_duration_clause': shard_duration_clause,
            })
        query = 'CREATE DATABASE {new_database_name}' + with_clause
        InfluxDBAdmin._execute_query(query, options)
        return True

    @staticmethod
    def create_retention_policy(
        policy_name,
        duration=None,
        replication=None,
        shard_duration=None,
        is_default=False,
    ):
        if not duration and not replication and not shard_duration\
           and not is_default:
            msg = '`duration` or `replication` or `shard_duration` ' +\
                  ' or `is_default` must be not null'
            raise exceptions.InfluxDBError(msg)

        if (not duration and replication) or (duration and not replication):
            msg = '`duration` or `replication` must be not null'
            raise exceptions.InfluxDBError(msg)

        policy_name = GenericDBAdminCommand._format_with_double_quote(
            policy_name,
        )

        default_clause = GenericDBAdminCommand._generate_default_clause(
            is_default
        )
        duration_clause = GenericDBAdminCommand._generate_duration_clause(
            duration
        )
        replication_clause = GenericDBAdminCommand._generate_replication_clause(
            replication
        )
        shard_duration_clause = GenericDBAdminCommand._generate_shard_duration_clause(
            shard_duration
        )

        options = {
            'default_clause': default_clause,
            'duration_clause': duration_clause,
            'replication_clause': replication_clause,
            'policy_name': policy_name,
            'shard_duration_clause': shard_duration_clause,
        }
        query = 'CREATE RETENTION POLICY {policy_name} ON {database_name}' +\
                ' {duration_clause}' +\
                ' {replication_clause}' +\
                ' {shard_duration_clause}' +\
                ' {default_clause}'
        InfluxDBAdmin._execute_query(query, options)
        return True

    @staticmethod
    def create_subscription(subscription_name, hosts, any=False):
        subscription_name = GenericDBAdminCommand._format_with_double_quote(
            subscription_name,
        )
        destination_type = 'ANY' if any else 'ALL'
        formatted_hosts = ', '.join(['\'{}\''.format(h) for h in hosts])
        options = {
            'hosts': formatted_hosts,
            'destination_type': destination_type,
            'subscription_name': subscription_name,
        }
        query = 'CREATE SUBSCRIPTION {subscription_name}' +\
                ' ON {full_database_name}' +\
                ' DESTINATIONS {destination_type} {hosts}'
        InfluxDBAdmin._execute_query(query, options)
        return True

    @staticmethod
    def create_user(user_name, password, with_privileges=False):
        user_name = GenericDBAdminCommand._get_formatted_user_name(user_name)
        password = GenericDBAdminCommand._format_with_simple_quote(
            password,
        )
        privilege_clause = 'WITH ALL PRIVILEGES' if with_privileges else ''
        options = {
            'user_name': user_name,
            'password': password,
            'privilege_clause': privilege_clause,
        }
        query = 'CREATE USER {user_name} WITH PASSWORD {password}' +\
                ' {privilege_clause}'
        InfluxDBAdmin._execute_query(query, options)
        return True


class DeleteAdminCommand:
    @staticmethod
    def delete(measurements=[], criteria=[]):
        from_clause = GenericDBAdminCommand._generate_from_clause(measurements)
        where_clause = GenericDBAdminCommand._generate_where_clause(criteria)

        if not from_clause and not where_clause:
            msg = '`measurements` or `criteria` must be not null'
            raise exceptions.InfluxDBError(msg)

        options = {
            'from_clause': from_clause,
            'where_clause': where_clause,
        }
        query = 'DELETE {from_clause} {where_clause}'
        InfluxDBAdmin._execute_query(query, options)
        return True


class DropAdminCommand:
    @staticmethod
    def drop_continuous_query(query_name):
        query_name = GenericDBAdminCommand._format_with_double_quote(
            query_name,
        )
        options = {'query_name': query_name}
        query = 'DROP CONTINUOUS QUERY {query_name} ON {database_name}'
        InfluxDBAdmin._execute_query(query, options)
        return True

    @staticmethod
    def drop_database(database_name_to_delete):
        _database_name = GenericDBAdminCommand._format_with_double_quote(
            database_name_to_delete,
        )
        options = {'_database_name': _database_name}
        query = 'DROP DATABASE {_database_name}'
        InfluxDBAdmin._execute_query(query, options)
        return True

    @staticmethod
    def drop_measurement(measurement_name):
        measurement_name = GenericDBAdminCommand._format_with_double_quote(
            measurement_name,
        )
        options = {'measurement_name': measurement_name}
        query = 'DROP MEASUREMENT {measurement_name}'
        InfluxDBAdmin._execute_query(query, options)
        return True

    @staticmethod
    def drop_retention_policy(policy_name):
        policy_name = GenericDBAdminCommand._format_with_double_quote(
            policy_name,
        )
        options = {'policy_name': policy_name}
        query = 'DROP RETENTION POLICY {policy_name} ON {database_name}'
        InfluxDBAdmin._execute_query(query, options)
        return True

    @staticmethod
    def drop_series(measurements=[], criteria=[]):
        from_clause = GenericDBAdminCommand._generate_from_clause(measurements)
        where_clause = GenericDBAdminCommand._generate_where_clause(criteria)

        if not from_clause and not where_clause:
            msg = '`measurements` or `criteria` must be not null'
            raise exceptions.InfluxDBError(msg)

        options = {
            'from_clause': from_clause,
            'where_clause': where_clause,
        }
        query = 'DROP SERIES {from_clause} {where_clause}'
        InfluxDBAdmin._execute_query(query, options)
        return True

    @staticmethod
    def drop_shard(shard_id):
        options = {'shard_id': shard_id}
        query = 'DROP SHARD {shard_id}'
        InfluxDBAdmin._execute_query(query, options)
        return True

    @staticmethod
    def drop_subscription(subscription_name):
        subscription_name = GenericDBAdminCommand._format_with_double_quote(
            subscription_name,
        )
        options = {'subscription_name': subscription_name}
        query = 'DROP SUBSCRIPTION {subscription_name} ON {full_database_name}'
        InfluxDBAdmin._execute_query(query, options)
        return True

    @staticmethod
    def drop_user(user_name):
        user_name = GenericDBAdminCommand._get_formatted_user_name(user_name)
        options = {'user_name': user_name}
        query = 'DROP USER {user_name}'
        InfluxDBAdmin._execute_query(query, options)
        return True


class ExplainAdminCommand:
    @staticmethod
    def explain(query, analyze=False):
        analyze = 'ANALYZE' if analyze else ''
        options = {
            'analyze': analyze,
            'query': query,
        }
        query = 'EXPLAIN {analyze} {query}'
        parser = serializers.FlatFormattedSerieSerializer
        return InfluxDBAdmin._execute_query_with_parser(query, parser, options)


class GrantAdminCommand:
    @staticmethod
    def grant(privilege, user_name):
        privilege = GenericDBAdminCommand._get_formatted_privilege(privilege)
        user_name = GenericDBAdminCommand._get_formatted_user_name(user_name)
        options = {
            'privilege': privilege,
            'user_name': user_name,
        }
        query = 'GRANT {privilege} ON {database_name} TO {user_name}'
        parser = serializers.FlatFormattedSerieSerializer
        return InfluxDBAdmin._execute_query_with_parser(query, parser, options)


class KillAdminCommand:
    @staticmethod
    def kill(query_id):
        options = {'query_id': query_id}
        query = 'KILL QUERY {query_id}'
        parser = serializers.FlatFormattedSerieSerializer
        return InfluxDBAdmin._execute_query_with_parser(query, parser, options)


class RevokeAdminCommand:
    @staticmethod
    def revoke(privilege, user_name):
        privilege = GenericDBAdminCommand._get_formatted_privilege(privilege)
        user_name = GenericDBAdminCommand._get_formatted_user_name(user_name)
        options = {
            'privilege': privilege,
            'user_name': user_name,
        }
        query = 'REVOKE {privilege} ON {database_name} FROM {user_name}'
        parser = serializers.FlatFormattedSerieSerializer
        return InfluxDBAdmin._execute_query_with_parser(query, parser, options)


class ShowAdminCommand:
    @staticmethod
    def show_field_key_cardinality(exact=False):
        exact = 'EXACT' if exact else ''
        options = {'exact': exact}
        query = 'SHOW FIELD KEY {exact} CARDINALITY'
        parser = serializers.FormattedSerieSerializer
        return InfluxDBAdmin._execute_query_with_parser(query, parser, options)

    @staticmethod
    def show_measurement_cardinality(exact=False):
        exact = 'EXACT' if exact else ''
        options = {'exact': exact}
        query = 'SHOW MEASUREMENT {exact} CARDINALITY'
        parser = serializers.FlatSingleValueSerializer
        return InfluxDBAdmin._execute_query_with_parser(query, parser, options)

    @staticmethod
    def show_series_cardinality(exact=False):
        exact = 'EXACT' if exact else ''
        options = {'exact': exact}
        query = 'SHOW SERIES {exact} CARDINALITY'
        parser = serializers.FlatSingleValueSerializer
        if exact:
            parser = serializers.FormattedSerieSerializer
        return InfluxDBAdmin._execute_query_with_parser(query, parser, options)

    @staticmethod
    def show_tag_key_cardinality(exact=False):
        exact = 'EXACT' if exact else ''
        options = {'exact': exact}
        query = 'SHOW TAG KEY {exact} CARDINALITY'
        parser = serializers.FormattedSerieSerializer
        return InfluxDBAdmin._execute_query_with_parser(query, parser, options)

    @staticmethod
    def show_tag_values_cardinality(key, exact=False):
        key_clause = 'KEY = "{key}"'.format(key=key)
        exact = 'EXACT' if exact else ''
        options = {
            'exact': exact,
            'key_clause': key_clause,
        }
        query = 'SHOW TAG VALUES {exact} CARDINALITY WITH {key_clause}'
        parser = serializers.FormattedSerieSerializer
        return InfluxDBAdmin._execute_query_with_parser(query, parser, options)

    @staticmethod
    def show_continuous_queries():
        query = 'SHOW CONTINUOUS QUERIES'
        parser = serializers.FormattedSerieSerializer
        return InfluxDBAdmin._execute_query_with_parser(query, parser)

    @staticmethod
    def show_diagnostics():
        query = 'SHOW DIAGNOSTICS'
        parser = serializers.FormattedSerieSerializer
        return InfluxDBAdmin._execute_query_with_parser(query, parser)

    @staticmethod
    def show_field_keys(measurements=[]):
        from_clause = GenericDBAdminCommand._generate_from_clause(measurements)
        options = {'from_clause': from_clause}
        query = 'SHOW FIELD KEYS {from_clause}'
        parser = serializers.FormattedSerieSerializer
        return InfluxDBAdmin._execute_query_with_parser(query, parser, options)

    @staticmethod
    def show_grants(user_name):
        options = {'user_name': user_name}
        query = 'SHOW GRANTS FOR {user_name}'
        parser = serializers.FormattedSerieSerializer
        return InfluxDBAdmin._execute_query_with_parser(query, parser, options)

    @staticmethod
    def show_databases():
        query = 'SHOW DATABASES'
        parser = serializers.FlatSimpleResultSerializer
        return InfluxDBAdmin._execute_query_with_parser(query, parser)

    @staticmethod
    def show_measurements(criteria=[]):
        where_clause = GenericDBAdminCommand._generate_where_clause(criteria)
        options = {'where_clause': where_clause}
        query = 'SHOW MEASUREMENTS {where_clause}'
        parser = serializers.FlatSimpleResultSerializer
        return InfluxDBAdmin._execute_query_with_parser(query, parser, options)

    @staticmethod
    def show_queries():
        query = 'SHOW QUERIES'
        parser = serializers.FlatFormattedSerieSerializer
        return InfluxDBAdmin._execute_query_with_parser(query, parser)

    @staticmethod
    def show_retention_policies():
        query = 'SHOW RETENTION POLICIES'
        parser = serializers.FlatFormattedSerieSerializer
        return InfluxDBAdmin._execute_query_with_parser(query, parser)

    @staticmethod
    def show_series(measurements=[], criteria=[], limit=None, offset=None):
        from_clause = GenericDBAdminCommand._generate_from_clause(measurements)
        where_clause = GenericDBAdminCommand._generate_where_clause(criteria)
        limit_clause = GenericDBAdminCommand._generate_limit_clause(limit)
        offset_clause = GenericDBAdminCommand._generate_offset_clause(offset)
        options = {
            'from_clause': from_clause,
            'where_clause': where_clause,
            'limit_clause': limit_clause,
            'offset_clause': offset_clause,
        }
        query = 'SHOW SERIES ON {database_name}' +\
                ' {from_clause}' +\
                ' {where_clause}' +\
                ' {limit_clause}' +\
                ' {offset_clause}'
        parser = serializers.FlatFormattedSerieSerializer
        return InfluxDBAdmin._execute_query_with_parser(query, parser, options)

    @staticmethod
    def show_stats():
        query = 'SHOW STATS'
        parser = serializers.FormattedSerieSerializer
        return InfluxDBAdmin._execute_query_with_parser(query, parser)

    @staticmethod
    def show_shards():
        query = 'SHOW SHARDS'
        parser = serializers.FormattedSerieSerializer
        return InfluxDBAdmin._execute_query_with_parser(query, parser)

    @staticmethod
    def show_shard_groups():
        query = 'SHOW SHARD GROUPS'
        parser = serializers.FlatFormattedSerieSerializer
        return InfluxDBAdmin._execute_query_with_parser(query, parser)

    @staticmethod
    def show_subscriptions():
        query = 'SHOW SUBSCRIPTIONS'
        parser = serializers.FormattedSerieSerializer
        return InfluxDBAdmin._execute_query_with_parser(query, parser)

    @staticmethod
    def show_tag_keys(measurements=[]):
        from_clause = GenericDBAdminCommand._generate_from_clause(measurements)
        options = {'from_clause': from_clause}
        query = 'SHOW TAG KEYS {from_clause}'
        parser = serializers.FormattedSerieSerializer
        return InfluxDBAdmin._execute_query_with_parser(query, parser, options)

    @staticmethod
    def show_tag_values(key, measurements=[]):
        key_clause = 'KEY = "{key}"'.format(key=key)
        from_clause = GenericDBAdminCommand._generate_from_clause(measurements)
        options = {
            'key_clause': key_clause,
            'from_clause': from_clause,
        }
        query = 'SHOW TAG VALUES {from_clause} WITH {key_clause}'
        parser = serializers.FormattedSerieSerializer
        return InfluxDBAdmin._execute_query_with_parser(query, parser, options)

    @staticmethod
    def show_users():
        query = 'SHOW USERS'
        parser = serializers.FlatFormattedSerieSerializer
        return InfluxDBAdmin._execute_query_with_parser(query, parser)


class InfluxDBAdmin(
    GenericDBAdminCommand,
    AlterAdminCommand,
    CreateAdminCommand,
    DeleteAdminCommand,
    DropAdminCommand,
    ExplainAdminCommand,
    GrantAdminCommand,
    KillAdminCommand,
    RevokeAdminCommand,
    ShowAdminCommand,
):
    pass



import pytest
from decimal import Decimal as D
from influxable import attributes, exceptions
from influxable.db import Query
from influxable.measurement import Measurement, MeasurementMeta
from influxable.serializers import MeasurementPointSerializer


class TestMeasurement:
    def create_measurement_class(self):
        class MySampleMeasurement(Measurement):
            measurement_name = 'mysamplemeasurement'
            time = attributes.TimestampFieldAttribute(precision="s")
            value = attributes.IntegerFieldAttribute()
        measurement_cls = MySampleMeasurement
        return measurement_cls

    def create_measurement_class_with_required(self):
        class MySampleMeasurement(Measurement):
            measurement_name = 'mysamplemeasurement'
            time = attributes.TimestampFieldAttribute()
            value = attributes.IntegerFieldAttribute(is_nullable=False)
        measurement_cls = MySampleMeasurement
        return measurement_cls

    def test_meta_check_type_success(self):
        measurement_cls = self.create_measurement_class()
        assert measurement_cls.__class__ == MeasurementMeta

    def test_meta_get_attribute_names_success(self):
        measurement_cls = self.create_measurement_class()
        attr_names = measurement_cls._get_attribute_names()
        assert attr_names == ['__attribute__time', '__attribute__value']

    def test_meta_check_has_query_attr_success(self):
        measurement_cls = self.create_measurement_class()
        assert hasattr(measurement_cls, 'get_query')

    def test_meta_create_instance_success(self):
        measurement_cls = self.create_measurement_class()
        instance = measurement_cls()
        assert isinstance(instance, measurement_cls)
        assert hasattr(instance, 'get_query')
        assert hasattr(instance, '_get_attributes')

    def test_factory_get_query_success(self):
        measurement_cls = self.create_measurement_class()
        query = measurement_cls.get_query()
        assert isinstance(query, Query)

    def test_meta_get_attributes_success(self):
        measurement_cls = self.create_measurement_class()
        attrs = measurement_cls._get_attributes()
        for attr in attrs:
            assert isinstance(attr, attributes.BaseAttribute)

    def test_meta_get_timestamp_attribute_success(self):
        measurement_cls = self.create_measurement_class()
        timestamp_attrs = measurement_cls._get_timestamp_attributes()
        for attr in timestamp_attrs:
            assert isinstance(attr, attributes.TimestampFieldAttribute)

    def test_measurement_name_success(self):
        measurement_cls = self.create_measurement_class()
        instance = measurement_cls()
        assert measurement_cls.measurement_name == 'mysamplemeasurement'
        assert instance.measurement_name == 'mysamplemeasurement'

    def test_parser_class_success(self):
        measurement_cls = self.create_measurement_class()
        instance = measurement_cls()
        assert measurement_cls.parser_class == MeasurementPointSerializer
        assert instance.parser_class == MeasurementPointSerializer

    def test_extend_attributes_success(self):
        measurement_cls = self.create_measurement_class()
        attrs = measurement_cls._get_attributes()
        for attr in attrs:
            attribute_name = attr.attribute_name
            ext_attribute_name = attr.ext_attribute_name
            assert hasattr(measurement_cls, attribute_name)
            assert hasattr(measurement_cls, ext_attribute_name)
            prop = getattr(measurement_cls, attribute_name)
            attr_field = getattr(measurement_cls, ext_attribute_name)
            assert isinstance(attr_field, attributes.BaseAttribute)
            assert isinstance(prop, property)

    def test_check_attributes_success(self):
        measurement_cls = self.create_measurement_class()
        instance = measurement_cls(time=1570481055, value=10)
        assert instance is not None

    def test_check_attributes_failed(self):
        with pytest.raises(exceptions.InfluxDBAttributeValueError):
            measurement_cls = self.create_measurement_class_with_required()
            measurement_cls()

    def test_clone_attributes_success(self):
        measurement_cls = self.create_measurement_class()
        instance = measurement_cls()
        cls_attrs = instance._get_attributes()
        for attr in cls_attrs:
            cloned_attribute = getattr(instance, attr.ext_attribute_name)
            assert attr != cloned_attribute

    def test_fill_values_success(self):
        measurement_cls = self.create_measurement_class()
        instance = measurement_cls(time=1570481055, value=10)
        assert instance.time == D(1570481055)
        assert instance.value == 10

    def test_fill_values_failed(self):
        with pytest.raises(exceptions.InfluxDBAttributeValueError):
            measurement_cls = self.create_measurement_class_with_required()
            measurement_cls(time='1570481055', value="S")

    def test_dict_success(self):
        measurement_cls = self.create_measurement_class()
        instance = measurement_cls(time=1570481055, value=10)
        assert instance.dict() == {'time': D('1570481055'), 'value': 10}

    def test_get_attributes_success(self):
        measurement_cls = self.create_measurement_class()
        instance = measurement_cls(time=1570481055, value=10)
        attrs = instance.get_attributes()
        for attr in attrs:
            assert isinstance(attr, attributes.BaseAttribute)

    def test_get_attribute_names_success(self):
        measurement_cls = self.create_measurement_class()
        instance = measurement_cls(time=1570481055, value=10)
        attr_names = instance.get_attribute_names()
        assert attr_names == ['time', 'value']

    def test_get_ext_attribute_names_success(self):
        measurement_cls = self.create_measurement_class()
        instance = measurement_cls(time=1570481055, value=10)
        attr_names = instance.get_ext_attribute_names()
        assert attr_names == ['__attribute__time', '__attribute__value']

    def test_get_timestamp_attributes_success(self):
        measurement_cls = self.create_measurement_class()
        instance = measurement_cls(time=1570481055, value=10)
        time_attrs = instance.get_timestamp_attributes()
        for attr in time_attrs:
            assert isinstance(attr, attributes.TimestampFieldAttribute)

    def test_get_prep_value_success(self):
        measurement_cls = self.create_measurement_class()
        instance = measurement_cls(time=1570481055, value=10)
        prep_value = instance.get_prep_value()
        assert prep_value == 'mysamplemeasurement value=10i 1570481055000000000'

    def test_items_success(self):
        measurement_cls = self.create_measurement_class()
        instance = measurement_cls(time=1570481055, value=10)
        items = {'time': D('1570481055'), 'value': 10}.items()
        assert instance.items() == items

    def test_bulk_success(self):
        measurement_cls = self.create_measurement_class()
        measurements = [
            measurement_cls(time=1570481055, value=10),
            measurement_cls(time=1570481065, value=20),
            measurement_cls(time=1570481075, value=30),
        ]
        res = measurement_cls.bulk_save(measurements)
        assert res is True

    def test_bulk_fail_1(self):
        with pytest.raises(exceptions.InfluxDBAttributeValueError):
            measurement_cls = self.create_measurement_class()
            measurements = True
            measurement_cls.bulk_save(measurements)

    def test_bulk_fail_2(self):
        with pytest.raises(exceptions.InfluxDBAttributeValueError):
            measurement_cls = self.create_measurement_class()
            measurements = [True, True, True]
            measurement_cls.bulk_save(measurements)



import requests
from urllib.parse import urljoin
from .decorators import raise_if_error


class InfluxDBRequest(requests.Session):
    def __init__(self, base_url, database_name, auth):
        super().__init__()
        self.base_url = base_url
        self.database_name = database_name
        self.auth = auth

    @raise_if_error
    def request(self, method, url, **kwargs):
        full_url = urljoin(self.base_url, url)
        return super().request(method, url=full_url, **kwargs)

    @raise_if_error
    def head(self, url, **kwargs):
        full_url = urljoin(self.base_url, url)
        return super().head(full_url, **kwargs)

    @raise_if_error
    def get(self, url, **kwargs):
        full_url = urljoin(self.base_url, url)
        return super().get(full_url, **kwargs)

    @raise_if_error
    def post(self, url, **kwargs):
        full_url = urljoin(self.base_url, url)
        return super().post(full_url, **kwargs)

    @raise_if_error
    def put(self, url, **kwargs):
        full_url = urljoin(self.base_url, url)
        return super().put(full_url, **kwargs)

    @raise_if_error
    def patch(self, url, **kwargs):
        full_url = urljoin(self.base_url, url)
        return super().patch(full_url, **kwargs)

    @raise_if_error
    def delete(self, url, **kwargs):
        full_url = urljoin(self.base_url, url)
        return super().delete(full_url, **kwargs)
